# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import sys
import re


def tag_MISC(dic, verbose=True):
    """Return dictionary with added tag within the value tuple."""
    out = {}
    for ind, (nsw, tag) in dic.items():
        if verbose:
            sys.stdout.write("\r{} of {} subtagged".format(len(out), len(dic)))
            sys.stdout.flush()
        if looks_rude(nsw):
            out.update({ind: (nsw, tag, 'PROF')})
        elif is_url(nsw):
            out.update({ind: (nsw, tag, 'URL')})
        elif hashtag_pattern.match(nsw):
            out.update({ind: (nsw, tag, 'HTAG')})
        else:
            out.update({ind: (nsw, tag, 'NONE')})
    if verbose:
        sys.stdout.write("\r{} of {} classified".format(len(out), len(dic)))
        sys.stdout.flush()
        print("\n")
    return out


def looks_rude(w):
    """Return 'True' if w is three + characters of only letters and '*'."""
    return len(w) > 2 and allrude(w)


def allrude(w):
    """Return 'True' if all characters are letters or '*'."""
    for lt in w:
        if not lt == '*' and not lt.isalpha():
            return False
    return True


def is_url(w):
    """Return 'True' if start or end of w looks like a url."""
    if urlstart_pattern.match(w) or urlend_pattern.match(w):
        return True
    else:
        return False

urlstart_pattern = re.compile('''
(https?://)|            #'http' followed by optional 's', then '://' OR
(www\.)                 #'www.'
''', re.VERBOSE | re.IGNORECASE)

urlend_pattern = re.compile('''
.*                      #any number of characters
\.                      # '.'
((com)|                 # 'com' OR
(org(\.uk)?)|           # 'org' followed optionally by '.uk' OR
(co\.uk))               # 'co.uk'
$                       #end of string
''', re.VERBOSE | re.IGNORECASE)

hashtag_pattern = re.compile('''
\#
[A-Za-z0-9]+
[_-]?
[A-Za-z0-9]*
''', re.VERBOSE)



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import re
import collections
from nltk.corpus import brown


def words(text):
    return re.findall('[a-z]+', text.lower())


def train(features):
    model = collections.defaultdict(lambda: 1)
    for f in features:
        model[f] += 1
    return model

NWORDS = train(brown.words())
alphabet = 'abcdefghijklmnopqrstuvwxyz'


def edits1(word):
    s = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [a + b[1:] for a, b in s if b]
    transposes = [a + b[1] + b[0] + b[2:] for a, b in s if len(b) > 1]
    replaces = [a + c + b[1:] for a, b in s for c in alphabet if b]
    inserts = [a + c + b for a, b in s for c in alphabet]
    return set(deletes + transposes + replaces + inserts)


def known_edits2(word):
    return set(e2 for e1 in edits1(word) for e2 in edits1(e1) if e2 in NWORDS)


def known(words):
    return set(w for w in words if w in NWORDS)


def correct(word):
    candidates = (known([word]) or
                  known(edits1(word)) or
                  known_edits2(word) or
                  [word])
    return max(candidates, key=NWORDS.get)



from sys import argv
import argparse
import pickle

from normalise import normalise, rejoin, tokenize_basic


def main():
    parser = argparse.ArgumentParser(description="""normalises text in file
     at path given (path/name.extension'), and stores in 'file_normalised' (path/name_normalised.extention). Will use
     simple default tokenizer and general abbrevations: to use custom tokenizer
     and abbreviations import normalise function from the module.""")
    parser.add_argument('text', metavar='P', type=str, nargs='+',
                        help='The path of the text to be normalised')
    parser.add_argument('-E', dest='english', type=str, choices=['BrE', 'AmE'],
                        default='BrE',
                        help='specify the variety as American English or British English (default: British English)')
    parser.add_argument('-A', dest='abbrevs', default=None,
                        help='The path to a pickled custom abbreviation dictionary')
    parser.add_argument('-T', dest='tokenizer', default=None,
                        help='The path to a pickled custom tokenizer')
    parser.add_argument('-V', '--verbose', dest='verbose', action='store_const',
                        const=True, default=False,
                        help='specify verbose output (default: False)')
    args = parser.parse_args()
    if args.abbrevs:
        with open(args.abbrevs, mode='rb') as f:
            user_abbrevs = pickle.load(f)
    else:
        user_abbrevs = {}
    if args.tokenizer:
        with open(args.tokenizer, mode='rb') as f:
            tokenizer = pickle.load(f)
    else:
        tokenizer = tokenize_basic
    for f in args.text:
        print(f)
        with open(f, mode='r') as raw:
            text = raw.read()
        i = f.rfind('.')
        with open('{}_normalised{}'.format(f[:i], f[i:]), mode='w') as out:
            norm = rejoin(normalise(text, verbose=args.verbose, variety=args.english, user_abbrevs=user_abbrevs, tokenizer=tokenizer))
            print(norm)
            out.write(norm)



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import sys
import re
import pickle
from io import open

import numpy as np
from sklearn.semi_supervised import LabelPropagation as lp
from roman import romanNumeralPattern

from normalise.detect import mod_path
from normalise.tagger import tagify, is_digbased, acr_pattern
from normalise.class_NUMB import gen_frame
from normalise.splitter import split, retagify
from normalise.data.measurements import meas_dict, meas_dict_pl
from normalise.data.abbrev_dict import abbrev_dict
from normalise.data.element_dict import element_dict

with open('{}/data/wordlist.pickle'.format(mod_path), mode='rb') as file:
    wordlist = pickle.load(file)

with open('{}/data/NSW_dict.pickle'.format(mod_path), mode='rb') as file:
    NSWs = pickle.load(file)

with open('{}/data/word_tokenized.pickle'.format(mod_path), mode='rb') as file:
    word_tokenized = pickle.load(file)

with open('{}/data/word_tokenized_lowered.pickle'.format(mod_path), mode='rb') as f:
    word_tokenized_lowered = pickle.load(f)

with open('{}/data/clf_ALPHA.pickle'.format(mod_path), mode='rb') as file:
    clf_ALPHA = pickle.load(file)

with open('{}/data/names.pickle'.format(mod_path), mode='rb') as file:
    names_lower = pickle.load(file)

if __name__ == "__main__":
    tagged = tagify(NSWs, verbose=False)

    ALPHA_dict = {ind: (nsw, tag) for ind, (nsw, tag) in tagged.items()
                  if tag == 'ALPHA'}

    SPLT_dict = {ind: (nsw, tag) for ind, (nsw, tag) in tagged.items()
                 if tag == 'SPLT'}

    splitted = split(SPLT_dict, verbose=False)
    retagged = retagify(splitted, verbose=False)
    retagged_ALPHA_dict = {ind: (nsw, tag)
                           for ind, (nsw, tag) in retagged.items()
                           if tag == 'SPLT-ALPHA'}
    ALPHA_dict.update(retagged_ALPHA_dict)

    ALPHAs_context = []
    for item in ALPHA_dict.items():
        if item[0] < 1206200:
            for word in gen_frame(item, word_tokenized):
                ALPHAs_context.append(' {} '.format(word))

    third_ALPHA_dict = {}
    count = 0
    for item in ALPHA_dict.items():
        count += 1
        if count % 3 == 0:
            third_ALPHA_dict.update((item,))

ampm = ['am', 'pm', 'AM', 'PM', 'a.m.', 'p.m.', 'A.M.', 'P.M.', 'pm.', 'am.']
adbc = ['AD', 'A.D.', 'ad', 'a.d.', 'BC', 'B.C.', 'bc', 'B.C.']


def run_clfALPHA(dic, text, verbose=True, user_abbrevs={}):
    """Train classifier on training data, return dictionary with added tag.

    dic: dictionary entry where key is index of word in orig text, value
         is a tuple with the nsw and the tag (ALPHA, NUMB, SPLT-, MISC).
    The dictionary returned has the same entries with the tuple extended with
    a more specific number tag assigned to it by the classifier.
    """

    clf = clf_ALPHA
    int_tag_dict = {
                    1: 'EXPN',
                    2: 'LSEQ',
                    3: 'WDLK',
                    }
    out = {}
    for (ind, (nsw, tag)) in dic.items():
        if verbose:
            sys.stdout.write("\r{} of {} classified".format(len(out), len(dic)))
            sys.stdout.flush()
        if romanNumeralPattern.match(nsw) and gen_frame((ind, (nsw, tag)), text)[1].lower() in names_lower:
            out.update({ind: (nsw, 'NUMB', 'NORD')})
        if nsw in user_abbrevs:
            out.update({ind: (nsw, 'ALPHA', 'EXPN')})
        else:
            pred_int = int(clf.predict(gen_featuresetsALPHA({ind: (nsw, tag)}, text)))
            ntag = int_tag_dict[pred_int]
            out.update({ind: (nsw, tag, ntag)})
    if verbose:
        sys.stdout.write("\r{} of {} classified".format(len(out), len(dic)))
        sys.stdout.flush()
        print("\n")
    return out


def gen_featuresetsALPHA(tagged_dict, text):
    """Return an array for features for each item in the input dict."""
    return np.array([give_featuresALPHA(item, text)
                    for item in tagged_dict.items()],
                    dtype=float)


def give_featuresALPHA(item, text):
    """Return a list of features for a dictionary item."""
    ind, nsw, tag = item[0], item[1][0], item[1][1]
    context = gen_frame(item, text)
    out = [

           ]
    out.extend(seed_features(item, context))
    """out.extend(in_features(nsw))"""
    return out


def seed_features(item, context):
    """Return a list of features equivalent to those used in the seedset."""
    ind, nsw, tag = item[0], item[1][0], item[1][1]
    out = [
           nsw in ['Mr.', 'Mrs.', 'Mr', 'Mrs'],
           nsw in ['i.e.', 'ie.', 'e.g.', 'eg.'],
           nsw.endswith('.') and nsw.istitle() and not acr_pattern.match(nsw),
           (nsw.isupper() and is_cons(nsw) and not (nsw in meas_dict
            and is_digbased(context[1])) and not acr_pattern.match(nsw)),
           (nsw in meas_dict or nsw in meas_dict_pl) and is_digbased(context[1]),
           (nsw in ampm or nsw in adbc) and is_digbased(context[1]),
           (nsw.istitle() and nsw.isalpha() and len(nsw) > 3 and not is_cons(nsw)),
           (((nsw.startswith("O'") or nsw.startswith("D'")) and nsw[2:].istitle())
           or (nsw.endswith("s'") and nsw[:-2].istitle())
           or (nsw.endswith("'s") and nsw[:-2].istitle())),
           (not (nsw.isupper() or nsw.endswith('s') and nsw[:-1].isupper())
            and (nsw.lower() in wordlist
            or (nsw[:-1].lower() in wordlist and nsw.endswith('s')))
            and nsw not in ampm),
           triple_rep(nsw) and len(nsw) > 3,
           bool(acr_pattern.match(nsw) and nsw not in meas_dict),
           nsw.isalpha() and nsw.islower() and len(nsw) > 3,
           nsw.endswith('s') and nsw[:-1].isupper(),
           nsw in element_dict,
           nsw.isalpha and nsw.islower() and len(nsw) > 2,
           nsw.lower() in abbrev_dict or nsw in ['St.', 'st.', 'St']
           ]
    return out


def gen_seed(dic, text):
    """Return a list of the (integer) labels assigned to the seedset."""
    seedset = []
    for ind, (nsw, tag) in dic.items():
        seedset.append(seed((ind, (nsw, tag)), text))
    return seedset


def gen_feats_and_seed(dic, text):
    """Return a tuple of arrays - the first of features, the second, labels."""
    seedset = []
    featset = []
    for ind, (nsw, tag) in dic.items():
        seedset.append(seed((ind, (nsw, tag)), text))
        featset.append(give_featuresALPHA((ind, (nsw, tag)), text))
    return (np.array(featset, dtype=int), np.array(seedset))


def fit_clf(dic, text):
    """Fit a Label Propogation classifier to the input dictionary."""
    model = lp(tol=0.01)
    X, y = gen_feats_and_seed(dic, text)
    model.fit(X, y)
    return model


def fit_and_store_clf(dic, text):
    """fit a Label Propogation classifier, and store in clf_ALPHA.pickle"""
    clf = fit_clf(dic, text)
    with open('{}/data/clf_ALPHA.pickle'.format(mod_path), 'wb') as file:
        pickle.dump(clf, file, protocol=2)


def seed(dict_tup, text):
    """Assign a seedset label to the input tuple.

    Generate seeds for the seedset by assigning integer labels to obvious
    cases. Where there is no obvious case, '-1' is returned.
    """
    ind, nsw, tag = dict_tup[0], dict_tup[1][0], dict_tup[1][1]
    context = gen_frame((ind, (nsw, tag)), text)
    if nsw in ['Mr.', 'Mrs.', 'Mr', 'Mrs']:
        return 3
    elif nsw in ['i.e.', 'ie.', 'e.g.', 'eg.']:
        return 2
    elif nsw.endswith('.') and nsw.istitle() and not acr_pattern.match(nsw):
        return 1
    elif nsw.lower() in abbrev_dict or nsw in ['St.', 'st.', 'St']:
        return 1
    elif (nsw.isupper() and is_cons(nsw) and not (nsw in meas_dict
          and is_digbased(context[1]))):
            return 2
    elif nsw.endswith('s') and nsw[:-1].isupper():
        return 2
    elif (nsw in meas_dict or nsw in meas_dict_pl) and is_digbased(context[1]):
        return 1
    elif (nsw in ampm or nsw in adbc) and is_digbased(context[1]):
        return 2
    elif nsw.istitle() and nsw.isalpha() and len(nsw) > 3 and not is_cons(nsw):
        return 3
    elif (((nsw.startswith("O'") or nsw.startswith("D'")) and nsw[2:].istitle())
           or (nsw.endswith("s'") and nsw[:-2].istitle())):
               return 3
    elif nsw in element_dict:
        return 1
    elif (not (nsw.isupper() or nsw.endswith('s') and nsw[:-1].isupper())
          and (nsw.lower() in wordlist
          or (nsw[:-1].lower() in wordlist and nsw.endswith('s')))
         and nsw not in ampm):
            return 3
    elif triple_rep(nsw) and len(nsw) > 3:
        return 3
    elif nsw.isalpha() and nsw.islower() and len(nsw) > 3:
        return 3
    elif acr_pattern.match(nsw) and nsw not in meas_dict:
        return 2
    elif len(nsw) == 1:
        return 2
    elif nsw.isalpha and nsw.islower() and len(nsw) > 2:
        return 3
    else:
        return -1


def is_cons(w):
    """Return True if no vowels in w."""
    for lt in w:
        if lt in ['A', 'E', 'I', 'O', 'U', 'a', 'e', 'i', 'o', 'u']:
            return False
    return True


def triple_rep(w):
    """Return 'True' if w has a letter repeated 3 times consecutively."""
    for i in range(len(w) - 2):
        if w[i] == w[i + 1] and w[i] == w[i + 2] and w[i].isalpha():
            return True
    return False



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import sys
import re
import pickle
from io import open

from normalise.detect import mod_path
from normalise.data.measurements import meas_dict

with open('{}/data/NSW_dict.pickle'.format(mod_path), mode='rb') as file:
    NSWs = pickle.load(file)

curr_list = ['£', '$', '€']
AlPHA_dict, NUMB_dict, MISC_dict = {}, {}, {}


def tagify(dic, verbose=True):
    """Return dictionary with added tag.

    dic: dictionary entry where key is index of word in orig text, value
         is the nsw
    The dictionary returned has the same keys with the values being a tuple
    with nsw and its assigned tag.
    """
    out = {}
    for ind, it in dic.items():
        if verbose:
            sys.stdout.write("\r{} of {} tagged".format(len(out), len(dic)))
            sys.stdout.flush()
        if len(it) > 100:
            out.update({ind: (it, 'MISC')})
        if is_digbased(it):
            out.update({ind: (it, 'NUMB')})
        elif (only_alpha(it) and
              (not mixedcase_pattern.match(it) or
               len(it) <= 3 or (it[-1] == 's' and not
               mixedcase_pattern.match(it[:-1])))):
                    out.update({ind: (it, 'ALPHA')})
        elif it in meas_dict:
            out.update({ind: (it, 'ALPHA')})
        elif is_url(it) or hashtag_pattern.match(it):
            out.update({ind: (it, 'MISC')})
        elif looks_splitty(it):
            out.update({ind: (it, 'SPLT')})
        else:
            out.update({ind: (it, 'MISC')})
    if verbose:
        sys.stdout.write("\r{} of {} tagged".format(len(out), len(dic)))
        sys.stdout.flush()
        print("\n")
    return out


def looks_splitty(w):
    """Return 'True' if w looks like it should be split."""
    if has_digit(w) and has_alpha(w):
        return True
    elif hyphen_pattern.match(w):
        return True
    elif (only_alpha(w) and not w[:2] == 'Mc'and not w[:3] == 'Mac' and
          mixedcase_pattern.match(w)):
        return True
    elif emph_pattern.match(w):
        return True
    else:
        return False


def is_digbased(w):
    """Return 'True' if w is based around digits."""
    if len(w) == 0:
        return False
    if w[-2:] in ["'s", "s'"]:
        return is_digbased(w[:-2])
    elif w[-3:] in ecurr_dict:
        return is_digbased(w[:-3])
    elif w[-2:] in ['st', 'nd', 'rd', 'th']:
        return is_digbased(w[:-2])
    elif w[-1].lower() in end_dict or w[-1] == 's':
        return is_digbased(w[:-1])
    elif w[0] in curr_list:
        return is_digbased(w[1:])
    elif w[0] == "'" and w[1:].isdigit():
        return True
    elif w[0] == '+':
        return is_digbased(w[1:])
    elif w[-1] == '+':
        return is_digbased(w[:-1])
    for lt in w:
        if not lt.isdigit() and lt not in ['/', '.', ',',
                                           '-', '–', '%',
                                           ':', "'", '"',
                                           "°", "—", "′"]:
            return False
    else:
        return True and has_digit(w)


def only_alpha(w):
    """Return 'True' if w is based on alphabetic characters."""
    if len(w) == 0:
        return False
    if w[-2:] in ["'s", "s'"]:
        return only_alpha(w[:-2])
    if is_acr(w):
        return True
    if w[-1] == '.':
        return only_alpha(w[:-1])
    if w[0] == "'":
        return only_alpha(w[1:])
    if w[-1] == "'":
        return only_alpha(w[:-1])
    if w[0] == '"' and w[-1] == '"':
        return only_alpha(w[1:-1])
    return (has_alpha(w) and bool(alpha_pattern.match(w)))


def is_digit(lt):
    """Return 'True' if input letter is a digit."""
    return lt.isdigit()


def is_alpha(lt):
    """Return 'True' if input letter is a letter."""
    return lt.isalpha()


def has_digit(w):
    """Return 'True' if w contains a digit."""
    if list(filter(is_digit, w)):
        return True
    else:
        return False


def has_alpha(w):
    """Return 'True' if w contains a letter."""
    if is_digbased(w):
        return False
    else:
        if list(filter(is_alpha, w)):
            return True
        else:
            return False


def is_acr(w):
    """Return 'True' if w is an acronym (alternates letters and '.')."""
    return bool(acr_pattern.match(w))


def is_url(w):
    """Return 'True' if start or end of w looks like a url."""
    if urlstart_pattern.match(w) or urlend_pattern.match(w):
        return True
    else:
        return False

hyphen_pattern = re.compile('''
(([\'\.]?                       # optional "'" or "."
[A-Za-z]                        # letter
[\'\.]?)                        # optional "'" or "."
+                               # lines 1-3 repeated one or more times
(-|–|—|/|\s|(-&-))              # followed by '-', '–', '—', '/', or '-&-'
)+                              # all of the above repeated 1+ times
([\'\.]?[A-Za-z][\.\']?)*       # optional final 'word' (same as lines 1-3)
$                               # end of string
    ''', re.VERBOSE)

mixedcase_pattern = re.compile('''
([A-Z]{2,}[a-z]) |    # 2 or more capitals, a lowercase OR
(.*[a-z][A-Z])        # any number of chars, a lowercase, an uppercase
''', re.VERBOSE)

emph_pattern = re.compile('''
[\*~<:]+                          # '*' or '~' repeated one or more times
([A-Za-z]+[-/\']?)+[A-Za-z]+    # optionally hyphenated 'word'
[\*~>:]+                          # '*' or '~' repeated one or more times
$                               # end of string
''', re.VERBOSE)

acr_pattern = re.compile('''
([A-Za-z]\.)+           # letter followed by '.' 1 or more times
[A-Za-z]?               # optional final letter
$                       # end of string
''', re.VERBOSE)

alpha_pattern = re.compile('''
([A-Za-z]+\'?)*         # (1 or more letter, optional apostrophe) repeated
$                       # end of string
''', re.VERBOSE)

urlstart_pattern = re.compile('''
(https?://)|            #'http' followed by optional 's', then '://' OR
(www\.)                 #'www.'
''', re.VERBOSE | re.IGNORECASE)

urlend_pattern = re.compile('''
.*                      #any number of characters
\.                      # '.'
((com)|                 # 'com' OR
(org(\.uk)?)|           # 'org' followed optionally by '.uk' OR
(co\.uk))               # 'co.uk'
$                       #end of string
''', re.VERBOSE | re.IGNORECASE)

hashtag_pattern = re.compile('''
\#
[A-Za-z0-9]+
[_-]?
[A-Za-z0-9]*
''', re.VERBOSE)

end_dict = {'k': 'thousand', 'm': 'million', 'b': 'billion'}
ecurr_dict = {
              "US$": "U S Dollar",
              "AFN": "Afghanistan Afghani",
              "ALL": "Albanian Lek",
              "DZD": "Algerian Dinar",
              "AOA": "Angolan Kwanza",
              "ARS": "Argentine Peso",
              "AMD": "Armenian Dram",
              "AWG": "Aruban Florin",
              "AUD": "Australian Dollar",
              "AZN": "Azerbaijan New Manat",
              "BSD": "Bahamian Dollar",
              "BHD": "Bahraini Dinar",
              "BDT": "Bangladeshi Taka",
              "BBD": "Barbados Dollar",
              "BYR": "Belarusian Ruble",
              "BZD": "Belize Dollar",
              "BMD": "Bermudian Dollar",
              "BTN": "Bhutan Ngultrum",
              "BOB": "Bolivian Boliviano",
              "BAM": "Bosnian Mark",
              "BWP": "Botswana Pula",
              "BRL": "Brazilian Real",
              "BND": "Brunei Dollar",
              "BGN": "Bulgarian Lev",
              "BIF": "Burundi Franc",
              "XOF": "CFA Franc BCEAO",
              "XAF": "CFA Franc BEAC",
              "XPF": "CFP Franc",
              "KHR": "Cambodian Riel",
              "CAD": "Canadian Dollar",
              "CHF": "Swiss Franc",
              "CVE": "Cape Verde Escudo",
              "KYD": "Cayman Islands Dollar",
              "CLP": "Chilean Peso",
              "CNY": "Chinese Yuan",
              "COP": "Colombian Peso",
              "KMF": "Comoros Franc",
              "CDF": "Congolese Franc",
              "CRC": "Costa Rican Colon",
              "HRK": "Croatian Kuna",
              "CUC": "Cuban Convertible Peso",
              "CUP": "Cuban Peso",
              "CYP": "Cyprus Pound",
              "CZK": "Czech Koruna",
              "DKK": "Danish Krone",
              "DJF": "Djibouti Franc",
              "DOP": "Dominican Republic Peso",
              "XCD": "East Caribbean Dollar",
              "EGP": "Egyptian Pound",
              "SVC": "El Salvador Colon",
              "EEK": "Estonian Kroon",
              "ETB": "Ethiopian Birr",
              "EUR": "Euro",
              "FKP": "Falkland Islands Pound",
              "FJD": "Fiji Dollar",
              "GMD": "Gambian Dalasi",
              "GEL": "Georgian Lari",
              "GHS": "Ghanaian New Cedi",
              "GIP": "Gibraltar Pound",
              "XAU": "Gold (oz)",
              "GTQ": "Guatemalan Quetzal",
              "GNF": "Guinea Franc",
              "GYD": "Guyanese Dollar",
              "HTG": "Haitian Gourde",
              "HNL": "Honduran Lempira",
              "HKD": "Hong Kong Dollar",
              "HUF": "Hungarian Forint",
              "ISK": "Iceland Krona",
              "INR": "Indian Rupee",
              "IDR": "Indonesian Rupiah",
              "IRR": "Iranian Rial",
              "IQD": "Iraqi Dinar",
              "ILS": "Israeli New Shekel",
              "JMD": "Jamaican Dollar",
              "JPY": "Japanese Yen",
              "JOD": "Jordanian Dinar",
              "KZT": "Kazakhstan Tenge",
              "KES": "Kenyan Shilling",
              "KWD": "Kuwaiti Dinar",
              "KGS": "Kyrgyzstanian Som",
              "LAK": "Lao Kip",
              "LVL": "Latvian Lats",
              "LBP": "Lebanese Pound",
              "LSL": "Lesotho Loti",
              "LRD": "Liberian Dollar",
              "LYD": "Libyan Dinar",
              "LTL": "Lithuanian Litas",
              "MOP": "Macau Pataca",
              "MKD": "Macedonian Denar",
              "MGA": "Malagasy Ariary",
              "MWK": "Malawi Kwacha",
              "MYR": "Malaysian Ringgit",
              "MVR": "Maldive Rufiyaa",
              "MTL": "Maltese Lira",
              "MRO": "Mauritanian Ouguiya",
              "MUR": "Mauritius Rupee",
              "MXN": "Mexican Peso",
              "MDL": "Moldovan Leu",
              "MNT": "Mongolian Tugrik",
              "MAD": "Moroccan Dirham",
              "MZN": "Mozambique New Metical",
              "MMK": "Myanmar Kyat",
              "ANG": "NL Antillian Guilder",
              "NAD": "Namibia Dollar",
              "NPR": "Nepalese Rupee",
              "NZD": "New Zealand Dollar",
              "NIO": "Nicaraguan Cordoba Oro",
              "NGN": "Nigerian Naira",
              "KPW": "North Korean Won",
              "NOK": "Norwegian Krone",
              "OMR": "Omani Rial",
              "PKR": "Pakistan Rupee",
              "PAB": "Panamanian Balboa",
              "PGK": "Papua New Guinea Kina",
              "PYG": "Paraguay Guarani",
              "PEN": "Peruvian Nuevo Sol",
              "PHP": "Philippine Peso ",
              "PLN": "Polish Zloty",
              "QAR": "Qatari Rial",
              "RON": "Romanian New Leu",
              "RUB": "Russian Rouble",
              "RWF": "Rwandan Franc",
              "WST": "Samoan Tala",
              "STD": "Sao Tome/Principe Dobra",
              "SAR": "Saudi Riyal",
              "RSD": "Serbian Dinar",
              "SCR": "Seychelles Rupee",
              "SLL": "Sierra Leone Leone",
              "XAG": "Silver oz",
              "SGD": "Singapore Dollar",
              "SKK": "Slovak Koruna",
              "SIT": "Slovenian Tolar",
              "SBD": "Solomon Islands Dollar",
              "SOS": "Somali Shilling",
              "ZAR": "South African Rand",
              "KRW": "South-Korean Won",
              "LKR": "Sri Lanka Rupee",
              "SHP": "St Helena Pound",
              "SDG": "Sudanese Pound",
              "SRD": "Suriname Dollar",
              "SZL": "Swaziland Lilangeni",
              "SEK": "Swedish Krona",
              "CHF": "Swiss Franc",
              "SYP": "Syrian Pound",
              "TWD": "Taiwan Dollar",
              "TZS": "Tanzanian Shilling",
              "THB": "Thai Baht",
              "TOP": "Tonga Pa'anga",
              "TTD": "Trinidad/Tobago Dollar",
              "TND": "Tunisian Dinar",
              "TRY": "Turkish New Lira",
              "TMM": "Turkmenistan Manat",
              "USD": "US Dollar",
              "UGX": "Uganda Shilling",
              "UAH": "Ukraine Hryvnia",
              "UYU": "Uruguayan Peso",
              "AED": "United Arab Emir Dirham",
              "VUV": "Vanuatu Vatu",
              "VEB": "Venezuelan Bolivar",
              "VND": "Vietnamese Dong",
              "YER": "Yemeni Rial",
              "ZMK": "Zambian Kwacha",
              "ZWD": "Zimbabwe Dollar",
              "£": "pound",
              "$": "dollar",
              "€": "euro"
              }



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import os
import sys
import pickle
from io import open

from nltk import WordNetLemmatizer
from nltk.corpus import words
from nltk.corpus import nps_chat
from nltk.corpus import brown
from nltk.corpus import names
from normalise.data.contraction_list import contractions
from normalise.data.tech_words import tech_words

mod_path = os.path.dirname(__file__)

with open('{}/data/wordlist.pickle'.format(mod_path), mode='rb') as file:
    wordlist = pickle.load(file)

with open('{}/data/fake_data.pickle'.format(mod_path), mode='rb') as file:
    fake_data = pickle.load(file)

if __name__ == '__main__':
    word_tokenized = brown.words() + nps_chat.words() + fake_data
    brown_lower = {w.lower() for w in brown.words()
                   if len(w) > 4 and w.isalpha()}
    names_lower = {w.lower() for w in names.words()}
    words_lower = {w.lower() for w in words.words('en') if len(w) > 1}
    wordlist = brown_lower | names_lower | words_lower | set(tech_words) | {'I', 'i', 'a', 'A'}
    word_tokenized_lowered = [w.lower() if w.lower() in wordlist
                              else w for w in word_tokenized]
    word_tokenized = list(word_tokenized)


# Conditions for identification of NSWs.
def cond1(w):
    """ Return word if its lower-cased form is not in the wordlist."""
    return w.lower() not in wordlist or w == 'US'


def cond2(w):
    """ Return word if its lemmatised form is not in the wordlist."""
    wnl = WordNetLemmatizer()
    return wnl.lemmatize(w.lower()) not in wordlist


def cond3(w):
    """ Return word if it is not single punctuation."""
    if len(w) > 2:
        return True
    bools = [l.isalnum() for l in w]
    if True in bools:
        return True
    else:
        return False


def cond4(w):
    """ Return word if its non-possessive form (with 's/s' removed) is not in
    the wordlist.
    """
    return (not ((w.endswith("'s") or w.endswith("s'")) and
            w.lower()[:-2] in wordlist))


def ident_NSW(w):
    """ Identify NSWs.

    Return word if it satisfies all four above conditions.
    """
    return (cond1(w) and cond2(w) and cond3(w) and cond4(w)
            and not (w.lower() in contractions))


def create_NSW_dict(text, verbose=True):
    "Create dictionary of NSWs in text: keys are indices, values NSWs"
    return {i: text[i] for i in range(len(text)) if ident_NSW(text[i])}



# -*- coding: utf-8 -*-

element_dict = {
                "Ac": "Actinium",
                "Ag": "Silver",
                "Al": "Aluminum",
                "Am": "Americium",
                "Ar": "Argon",
                "As": "Arsenic",
                "At": "Astatine",
                "Au": "Gold",
                "Ba": "Barium",
                "Be": "Beryllium",
                "Bh": "Bohrium",
                "Bi": "Bismuth",
                "Bk": "Berkelium",
                "Br": "Bromine",
                "Ca": "Calcium",
                "Cd": "Cadmium",
                "Ce": "Cerium",
                "Cf": "Californium",
                "Cl": "Chlorine",
                "Cm": "Curium",
                "Co": "Cobalt",
                "Cr": "Chromium",
                "Cs": "Cesium",
                "Cu": "Copper",
                "Db": "Dubnium",
                "Ds": "Darmstadtium",
                "Dy": "Dysprosium",
                "Er": "Erbium",
                "Es": "Einsteinium",
                "Eu": "Europium",
                "Fe": "Iron",
                "Fm": "Fermium",
                "Fr": "Francium",
                "Ga": "Gallium",
                "Gd": "Gadolinium",
                "Ge": "Germanium",
                "H": "Hydrogen",
                "He": "Helium",
                "Hf": "Hafnium",
                "Hg": "Mercury",
                "Ho": "Holmium",
                "Hs": "Hassium",
                "In": "Indium",
                "Ir": "Iridium",
                "Kr": "Krypton",
                "La": "Lanthanum",
                "Li": "Lithium",
                "Lr": "Lawrencium",
                "Lu": "Lutetium",
                "Md": "Mendelevium",
                "Mg": "Magnesium",
                "Mn": "Manganese",
                "Mo": "Molybdenum",
                "Mt": "Meitnerium",
                "Na": "Sodium",
                "Nb": "Niobium",
                "Nd": "Neodymium",
                "Ne": "Neon",
                "Ni": "Nickel",
                "No": "Nobelium",
                "Np": "Neptunium",
                "Os": "Osmium",
                "Pa": "Protactinium",
                "Pb": "Lead",
                "Pd": "Palladium",
                "Pm": "Promethium",
                "Po": "Polonium",
                "Pr": "Praseodymium",
                "Pt": "Platinum",
                "Pu": "Plutonium",
                "Ra": "Radium",
                "Rb": "Rubidium",
                "Re": "Rhenium",
                "Rf": "Rutherfordium",
                "Rg": "Roentgenium",
                "Rh": "Rhodium",
                "Rn": "Radon",
                "Ru": "Ruthenium",
                "Sb": "Antimony",
                "Sc": "Scandium",
                "Se": "Selenium",
                "Sg": "Seaborgium",
                "Si": "Silicon",
                "Sm": "Samarium",
                "Sn": "Tin",
                "Sr": "Strontium",
                "Ta": "Tantalum",
                "Tb": "Terbium",
                "Tc": "Technetium",
                "Te": "Tellurium",
                "Th": "Thorium",
                "Ti": "Titanium",
                "Tl": "Thallium",
                "Tm": "Thulium",
                "Uub": "Ununbium",
                "Uuh": "Ununhexium",
                "Uuo": "Ununoctium",
                "Uup": "Ununpentium",
                "Uuq": "Ununquadium",
                "Uus": "Ununseptium",
                "Uut": "Ununtrium",
                "Xe": "Xenon",
                "Yb": "Ytterbium",
                "Zn": "Zinc",
                "Zr": "Zirconium",
                }



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

from math import log
from collections import defaultdict
import re
import pickle
from io import open

from nltk.corpus import wordnet as wn
from nltk.corpus import stopwords
from nltk.corpus import names
from nltk.tokenize import word_tokenize as wt
from nltk import FreqDist as fd
from nltk import pos_tag

from normalise.detect import mod_path
from normalise.data.abbrev_dict import states, create_user_abbrevs
from normalise.splitter import split
from normalise.tagger import is_digbased
from normalise.data.measurements import meas_dict, meas_dict_pl

with open('{}/data/word_tokenized_lowered.pickle'.format(mod_path), mode='rb') as f:
    word_tokenized_lowered = pickle.load(f)

with open('{}/data/pos_dicts.pickle'.format(mod_path), mode='rb') as file:
    pos_tag_dict, pos_tag_dict_univ = pickle.load(file)

with open('{}/data/abbrev_dict.pickle'.format(mod_path), mode='rb') as file:
    abbrevs_orig = pickle.load(file)

with open('{}/data/sig_dict.pickle'.format(mod_path), mode='rb') as file:
    sig_dict = pickle.load(file)

brown = word_tokenized_lowered[:1161192]
brown_common = {word: log(1161192 / freq) for
                word, freq in fd(brown).most_common(5000)[100:]}
words = [w for w, freq in fd(brown).most_common()]
names_lower = {w.lower() for w in names.words()}


def expand_EXPN(nsw, i, text, user_abbrevs={}):
    """Expand abbreviations to best possible match. If no close matches,
       return nsw."""
    try:
        if user_abbrevs:
            abbrevs = create_user_abbrevs(user_abbrevs)
        else:
            abbrevs = abbrevs_orig   
        if nsw in ['St.', 'st.', 'St']:
            if i < len(text):
                if text[i + 1].lower() in names_lower:
                    return 'Saint'
                elif text[i + 1].endswith("'s"):
                    if text[i + 1][:-2].lower() in names_lower:
                           return 'Saint'
                elif text[i - 1].istitle():
                    return 'street'
                elif text[i + 1].istitle():
                    return 'Saint'
        elif nsw in meas_dict:
            if isinstance(i, int):
                if is_digbased(text[i - 1]):
                    if text[i - 1] == '1':
                        return meas_dict[nsw]
                    else:
                        return meas_dict_pl[nsw]
            else:
                full = text[int(i)]
                index = full.find(nsw)
                if index == 0:
                    if is_digbased(text[int(i) - 1]):
                        if text[int(i) - 1] == '1':
                            return meas_dict[nsw]
                        else:
                            return meas_dict_pl[nsw]
                else:
                    if is_digbased(full[:index]):
                        if text[int(i) - 1] == '1':
                            return meas_dict[nsw]
                        else:
                            return meas_dict_pl[nsw]
        elif (nsw.endswith('.') and nsw[:-1] in meas_dict
              and is_digbased(text[i - 1])):
            if text[i - 1] == '1':
                return meas_dict[nsw[:-1]]
            else:
                return meas_dict_pl[nsw[:-1]]
        if nsw.endswith('.') and nsw[:-1].lower() in abbrevs:
            w = nsw[:-1]
        else:
            w = nsw
        if w.lower() in abbrevs:
            cands = abbrevs[w.lower()]
            true_tag = abbrev_tag(i, text)
            true_tag_univ = abbrev_tag_univ(i, text)
            if len(cands) == 1:
                cand = cands[0]
                if pos_tag_dict_univ[cand.lower()] in [true_tag_univ, tuple()]:
                    return cand
            matches = []
            for cand in cands:
                if true_tag in pos_tag_dict[cand.lower()]:
                    matches += [cand]
            if not matches:
                for cand in cands:
                    if true_tag_univ in pos_tag_dict_univ[cand.lower()]:
                        matches += [cand]
            if matches:
                best = 0
                current = []
                if len(matches) == 1:
                    return matches[0]
                for cand in matches:
                    olap = overlap(i, cand, text)
                    if olap > best and cand in brown_common:
                        best = olap
                        current = [cand]
                    elif olap == best and best != 0:
                        current.append(cand)
                    elif cand in states.values() and not current:
                        current.append(cand)
                best = 0
                exp = ''
                for c in current:
                    if c in states.values():
                        return c
                    elif c in brown_common:
                        freq = brown_common[c]
                    else:
                        freq = 0
                    if freq < best:
                        best = freq
                        exp = c
            else:
                exp = maximum_overlap(w, i, text)
        elif w.lower().endswith('s.') and w.lower()[:-2] in abbrevs:
            return expand_EXPN(w.lower()[:-2], i, text) + 's'
        elif w.lower().endswith('s') and w.lower()[:-1] in abbrevs:
            return expand_EXPN(w.lower()[:-1], i, text) + 's'
        else:
            exp = maximum_overlap(w, i, text)
        if exp == '':
            return nsw
        else:
            return exp
    except(KeyboardInterrupt, SystemExit):
        raise
    except:
        return nsw


def maximum_overlap(w, i, text):
    """Return the candidate expansion with the highest overlap."""
    best = 0
    current = []
    curr = ''
    t_matches = tag_matches(i, text)
    if t_matches:
        if len(t_matches) == 1:
            if t_matches[0] in brown_common:
                return t_matches[0]
            else:
                return w
        for cand in t_matches:
            olap = overlap(i, cand, text)
            if olap > best and cand in words:
                best = olap
                current = [cand]
            elif olap == best and best != 0:
                current.append(cand)
        best = 0
        for c in current:
            if c in brown_common:
                freq = brown_common[c]
            else:
                freq = 0
            if freq < best:
                best = freq
                curr = c
            elif freq == best and len(tag_matches(i, text)) == 1:
                best = freq
                curr = c
            return curr
    if curr == '':
        return w
    else:
        return curr


def overlap(i, word, text):
    """Return overlap between words in the context of the abbreviation and
       words in the signatures generated for each candidate expansion."""
    overlap = 0
    sig = gen_signature(word)
    context = gen_context(i, text)
    for w in context:
        if w in sig:
            if w in brown_common:
                overlap += brown_common[w]
            else:
                overlap += log(1161192 / 1)
    return overlap


def find_matches(word):
    """Find examples of the candidate word in the Brown corpus."""
    lst1 = []
    for i in range(len(brown)):
        if brown[i] == word:
            lst1.append(i)
    return lst1


def gen_signature(word):
    """Generate a signature for each candidate expansion, using contextual
       information from the Brown corpus, as well as WordNet definitions and
       examples (if applicable)."""
    if word in gen_signature.dict:
        return gen_signature.dict[word]
    inds = find_matches(word)
    if len(inds) > 50:
        f = len(inds) / 50
        inds = [inds[int(i * f)] for i in range(50)]
    signature = defaultdict(int)
    for i in inds:
        for w in gen_context(i, brown):
            signature[w] += 1
    sig = {w for w in signature
           if signature[w] > 1
           and w not in stopwords.words('english') and w != ','}
    if word in wn.words():
        if wn.synsets(word) and str(wn.synsets(word)[0]).count("'") == 2:
            define = (eval("wn.{}.definition()".format(
                      str(wn.synsets(word)[0]).lower())))
            examples = (eval("wn.{}.examples()".format(
                        str(wn.synsets(word)[0]).lower())))
            if examples:
                for ex in examples:
                        sig.update([w for w in wt(ex)
                                   if w not in stopwords.words('english')])
            if define:
                        sig.update([w for w in wt(define)
                                   if w not in stopwords.words('english')])
    gen_signature.dict[word] = sig
    return sig

gen_signature.dict = sig_dict

def gen_context(i, text):
    """Generate context for the abbreviation - 4 words either side unless
       sentence is too short."""
    ind = i
    context = []
    text = text[:]
    if not isinstance(i, int):
        ind = int(i)
        split_token = text[ind]
        del text[ind]
        parts = split({ind: (split_token, 'SPLT')}, verbose=False)
        for it in sorted(parts, reverse=True):
            text.insert(ind, parts[it][0])
    start = ind
    end = ind + 1
    sloop = True
    while sloop and start > 0:
        if text[start - 1] not in ['.', '!', '?']:
            start -= 1
        else:
            sloop = False
    eloop = True
    while eloop and end <= len(text) - 1:
        if text[end] in ['.', '!', '?']:
            eloop = False
        else:
            end += 1
    if ind - start < 4:
        if end - start >= 9:
            context += text[start: start + 9]
        else:
            context += text[start: end]
    elif end - ind < 5:
        if end - start >= 9:
            context += text[end - 9: end]
        else:
            context += text[start: end]
    else:
        context += text[ind - 4: ind + 5]
    return context


def tag_sent(i, text):
    """POS tag sentence (or context) containing abbreviation."""
    sent = gen_context(i, text)
    return pos_tag(sent)


def tag_cands(abbrv):
    """Tags candidate expansions of the abbreviaiton with POS."""
    tagged_cands = []
    for (cand, freq) in gen_best(abbrv):
        tagged_cands += [(cand, pos_tag_dict[cand])]
    return tagged_cands


def abbrev_tag(i, text):
    """Return POS tag for the abbreviation."""
    for (cand, tag) in tag_sent(i, text):
        if isinstance(i, int):
            if text[i] == cand:
                return tag
        else:
            if split({int(i): (text[int(i)], 'SPLT')},
                     verbose=False)[i][0] == cand:
                return tag


def tag_sent_univ(i, text):
    """POS tag sentence using universal tagset."""
    sent = gen_context(i, text)
    return pos_tag(sent, tagset='universal')


def tag_cands_univ(abbrv):
    """Tags candidate expansions using universal tagset."""
    tagged_cands = []
    for (cand, freq) in gen_best(abbrv):
        tagged_cands += [(cand, pos_tag_dict_univ[cand])]
    return tagged_cands


def abbrev_tag_univ(i, text):
    """Return POS tag for the abbreviation using universal tagset."""
    for (cand, tag) in tag_sent_univ(i, text):
        if isinstance(i, int):
            if text[i] == cand:
                return tag
        else:
            if split({int(i): (text[int(i)], 'SPLT')},
                     verbose=False)[i][0] == cand:
                return tag


def tag_matches(i, text):
    """Return candidate expansions whose POS tag matches the POS tag of the
       abbreviation."""
    matches = []
    if isinstance(i, int):
        abbrev = text[i]
    else:
        abbrev = split({int(i): (text[int(i)], 'SPLT')}, verbose=False)[i][0]
    true_tag = abbrev_tag(i, text)
    for (cand, tags) in tag_cands(abbrev):
        if true_tag in tags:
            matches += [cand]
    if not matches:
        true_tag_univ = abbrev_tag_univ(i, text)
        for (cand, tags) in tag_cands_univ(abbrev):
            if true_tag_univ in tags:
                matches += [cand]
    if not matches and len(tag_cands_univ(abbrev)) == 1:
        if tag_cands_univ(abbrev)[0][1] == tuple():
            return [tag_cands_univ(abbrev)[0][0]]
    if len(matches) <= 10:
        return matches
    else:
        return matches[:10]


def find_last_letter(w):
    """Find last alphabetic character in a word."""
    if w[-1].isalpha():
        return w[-1]
    else:
        return find_last_letter(w[:-1])


def gen_candidates(word):
    """Generate a list of candidate expansions given an abbreviation."""
    vowel_cands = []
    start_cands = []
    start_and_end_cands = []
    reg_cons = ''
    reg_start = ''
    reg_start_and_end = ''
    for lt in word.lower():
        if lt.isalpha():
            reg_cons += lt
            reg_cons += '[aeiou]*'
    reg_cons += '$'
    regex_cons = re.compile(reg_cons)
    for lt in word.lower():
        if lt.isalpha():
            reg_start += lt
    regex_start = re.compile(reg_start)
    last = find_last_letter(word)
    if last == 's':
        last = find_last_letter(word[:word.rfind(last)]) + last
    for lt in word[:word.rfind(last)].lower():
        if lt.isalpha():
            reg_start_and_end += lt
    reg_start_and_end += '.*{}$'.format(last)
    regex_start_and_end = re.compile(reg_start_and_end)
    for w in words:
        if regex_cons.match(w):
            vowel_cands.append(w)
        elif regex_start_and_end.match(w):
            start_and_end_cands.append(w)
        elif regex_start.match(w):
            start_cands.append(w)
    return vowel_cands, start_and_end_cands, start_cands


def distance(abbrv, word):
    """Calculate distance between abbreviation and potential expansion."""
    extras = [lt for lt in word if abbrv.count(lt) != word.count(lt)]
    count = 0
    for lt in extras:
        if lt not in ['a', 'e', 'i', 'o', 'u']:
            count += 1
        else:
            count += 0.2
    return count


def gen_best(abbrv):
    """Generate best expansions given abbreviation (up to 50)."""
    vowel_cands, start_and_end_cands, start_cands = gen_candidates(abbrv)
    vowel_freqs = sorted([(it, brown_common[it])
                         for it in vowel_cands
                         if it in brown_common],
                         key=lambda cand: cand[1])
    start_and_end_freqs = sorted([(it, brown_common[it])
                                 for it in start_and_end_cands
                                 if it in brown_common],
                                 key=lambda cand: cand[1])
    start_freqs = sorted([(it, brown_common[it])
                         for it in start_cands
                         if it in brown_common],
                         key=lambda cand: cand[1])
    ordered_cands = [(word, freq) for word, freq in vowel_freqs]
    start_and_end_ind = 0
    start_ind = 0
    while (start_and_end_ind < len(start_and_end_freqs) - 1
           and start_ind < len(start_freqs)):
        ordered_cands += [start_and_end_freqs[start_and_end_ind]]
        start_and_end_ind += 1
        ordered_cands += [start_and_end_freqs[start_and_end_ind]]
        start_and_end_ind += 1
        ordered_cands += [start_freqs[start_ind]]
        start_ind += 1
    if start_and_end_ind < len(start_and_end_freqs):
        ordered_cands.extend([(word, freq) for word, freq
                             in start_and_end_freqs[start_and_end_ind:]])
    if start_ind < len(start_freqs):
        ordered_cands.extend([(word, freq) for word, freq
                             in start_freqs[start_ind:]])
    for word in start_and_end_cands + start_cands:
        if word not in [w for w, f in ordered_cands]:
            ordered_cands += [(word, 0)]
    if len(ordered_cands) <= 50:
        return ordered_cands
    else:
        return ordered_cands[:50]



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import re
from io import open

from roman import romanNumeralPattern, fromRoman
from normalise.class_NUMB import gen_frame


def expand_fraction(n):
    """Expand fractions."""
    try:
        slash = n.find('/')
        first = n[:slash]
        second = n[slash + 1:]
        exp = ''
        if first.isdigit() and second.isdigit():
            if n == "1/2":
                exp = 'half'
            elif second == '100':
                if first == '1':
                    exp += "one hundredth"
                else:
                    exp += expand_NUM(first) + " hundredths"
            elif ((int(first) >= int(second)) or int(second) > 19
                    or second in ['1', '2']):
                exp += expand_NUM(first) + " over " + expand_NUM(second)
            else:
                numbers = ['3', '4', '5', '6', '7', '8', '9', '10', '11', '12',
                           '13', '14', '15', '16', '17', '18', '19']
                fractions = ['third', 'quarter', 'fifth', 'sixth', 'seventh',
                             'eighth', 'ninth', 'tenth', 'eleventh', 'twelfth',
                             'thirteenth', 'fourteenth', 'fifteenth',
                             'sixteenth', 'seventeenth', 'eighteenth',
                             'nineteenth']
                if first == '1':
                    exp += "one " + fractions[numbers.index(second)]
                else:
                    exp += (expand_NUM(first) + " "
                           + fractions[numbers.index(second)] + "s")
        else:
            return n
        return exp
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return n


def expand_NUM(n):
    """Return n as an cardinal in words."""
    try:
        if '/' in n:
            if whole_and_fract_pattern.match(n):
                m = whole_and_fract_pattern.match(n)
                exp = ''
                exp += expand_NUM(m.group(1))
                if m.group(2) == '1/2':
                    exp += " and a "
                else:
                    exp += " and "
                exp += expand_fraction(m.group(2))
                return exp
            else:
                return expand_fraction(n)

        if len(n) > 0:
            if n[-1] == 's':
                if len(n) > 1 and n[-2:] == "'s":
                    str = ''
                    str += expand_NUM(n[:-2])
                    str += "'s"
                else:
                    str = ''
                    if expand_NUM(n[:-1])[-1] == 'y':
                        str += expand_NUM(n[:-1])[:-1]
                        str += 'ies'
                    else:
                        str += expand_NUM(n[:-1])
                        str += 's'
                return str

        if dec2_pattern.match(n):
            str2 = ''
            str2 += (expand_NUM(dec2_pattern.match(n).group(1)) + " "
                     + expand_NUM(dec2_pattern.match(n).group(2)))
            return str2

        if n.startswith('.'):
            return "point " + expand_NDIG(n[1:])

        ones_C = [
                 "zero", "one", "two", "three", "four", "five", "six", "seven",
                 "eight", "nine", "ten", "eleven", "twelve", "thirteen",
                 "fourteen", "fifteen", "sixteen", "seventeen", "eighteen",
                 "nineteen"
                 ]

        tens_C = [
                 "zero", "ten", "twenty", "thirty", "forty", "fifty", "sixty",
                 "seventy", "eighty", "ninety"
                 ]

        large = [
                  "", "thousand", "million", "billion", "trillion", "quadrillion",
                  "quintillion", "sextillion", "septillion", "octillion",
                  "nonillion", "decillion", "undecillion", "duodecillion",
                  "tredecillion", "quattuordecillion", "sexdecillion",
                  "septendecillion", "octodecillion", "novemdecillion",
                  "vigintillion"
                  ]

        def subThousand(n):
            """Convert a cardinal to words for numbers less than a thousand."""
            if n <= 19:
                return ones_C[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                return tens_C[q] + (" " + subThousand(r) if r else "")
            else:
                q, r = divmod(n, 100)
                return (ones_C[q]
                        + " hundred"
                        + (" and " + subThousand(r) if r else ""))

        def splitByThousands(n):
            "Return reversed digits of n in groups of 3."""
            res = []
            while n:
                n, r = divmod(n, 1000)
                res.append(r)
            return res

        def thousandUp(n):
            """Return cardinal greater than a thousand in words."""
            list1 = []
            for i, z in enumerate(splitByThousands(n)):
                if i and z:
                    list1.insert(0, subThousand(z) + " " + large[i])
                elif z:
                        list1.insert(0, subThousand(z))
            return ", ".join(list1)

        def decimal(n):
            """Returns pronounced words with n as rhs of a decimal"""
            if n == '00':
                out = ''
            else:
                out = ' point'
                for lt in n:
                    out += ' {}'.format(ones_C[int(lt)])
            return out

        n_clean = ''
        for i in range(len(n)):
            if not i:
                if n[i].isdigit() or n[i] == '-':
                    n_clean += n[i]
                elif n[i] == '+':
                    return "plus " + expand_NUM(n[1:])
            else:
                if n[i].isdigit() or n[i] == '.':
                    n_clean += n[i]
        if '.' in n_clean:
            dot = n_clean.find('.')
            whole, part = n_clean[:dot], n_clean[dot + 1:]
            return expand_NUM(whole) + decimal(part)
        num = int(n_clean)
        if num == 0:
            return "zero"
        else:
            w = ("minus " if num < 0 else "") + thousandUp(abs(num))
            if len(n) < 4:
                return w
            else:
                if n[-3] == '0' and n[-1] != '0':
                    ind = w.rfind(",")
                    return w[:ind] + " and" + w[ind + 1:]
                else:
                    return w

    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return n


def expand_NRANGE(n):
    """Return a number range in words."""
    try:
        if range_pattern.match(n):
            m = range_pattern.match(n)
            str = ''
            if m.group(1).isdigit() and m.group(3).isdigit():
                if 1800 <= int(m.group(1)) < 2050:
                    str += expand_NYER(m.group(1))
                else:
                    str += expand_NUM(m.group(1))
            else:
                str += expand_NUM(m.group(1))
            if m.group(2) in ['/', '-', '–']:
                str += ' to '
                if m.group(1).isdigit() and m.group(3).isdigit():
                    if 1800 <= int(m.group(3)) < 2050:
                        str += expand_NYER(m.group(3))
                    else:
                        str += expand_NUM(m.group(3))
                else:
                    str += expand_NUM(m.group(3))
            return str
        elif ',' in n:
            hyph = n.find('-')
            one = n[:hyph]
            two = n[hyph + 1:]
            str = ''
            str += expand_NUM(one) + " to " + expand_NUM(two)
            return str
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return n


def expand_NORD(dict_tup, text):
    """Insert 'the' and 'of' if ordinal is a date, to sound more natural.
       Only if not already in text, eg. 'The 30th of April'."""
    try:
        ind, (nsw, tag, ntag) = dict_tup
        if romanNumeralPattern.match(nsw):
            return 'the ' + expand_NORD((ind, (str(fromRoman(nsw)), tag, ntag)), text)
        out = expand_ordinal(nsw)
        if gen_frame(dict_tup, text)[3] in months:
            out += ' of'
            if gen_frame(dict_tup, text)[1] not in ['The', 'the', 'A', 'a']:
                out = 'the ' + out
        return out
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return dict_tup[1][0]


def expand_ordinal(n):
    try:
        """Return n as an ordinal in words."""
        ones_C = [
                 "zero", "one", "two", "three", "four", "five", "six", "seven",
                 "eight", "nine", "ten", "eleven", "twelve", "thirteen",
                 "fourteen", "fifteen", "sixteen", "seventeen", "eighteen",
                 "nineteen"
                 ]

        ones_O = [
                  "zero", "first", "second", "third", "fourth", "fifth", "sixth",
                  "seventh", "eighth", "ninth", "tenth", "eleventh", "twelfth",
                  "thirteenth", "fourteenth", "fifteenth", "sixteenth",
                  "seventeenth", "eighteenth", "nineteenth"
                  ]

        tens_C = [
                  "zero", "ten", "twenty", "thirty", "forty", "fifty", "sixty",
                  "seventy", "eighty", "ninety"
                  ]

        tens_O = [
                  "zero", "tenth", "twentieth", "thirtieth", "fortieth",
                  "fiftieth", "sixtieth", "seventieth", "eightieth", "ninetieth"
                  ]

        large = [
                 "", "thousand", "million", "billion", "trillion", "quadrillion",
                 "quintillion", "sextillion", "septillion", "octillion",
                 "nonillion", "decillion", "undecillion", "duodecillion",
                 "tredecillion", "quattuordecillion", "sexdecillion",
                 "septendecillion", "octodecillion", "novemdecillion",
                 "vigintillion"
                 ]

        def subThousand_O(n):
            """Convert an ordinal to words for numbers less than a thousand."""
            if n <= 19:
                return ones_O[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                if r == 0:
                    return tens_O[q]
                else:
                    return tens_C[q] + "-" + ones_O[r]
            else:
                q, r = divmod(n, 100)
                if r == 0:
                    return ones_C[q] + " hundredth"
                else:
                    return ones_C[q] + " hundred" + " and " + subThousand_O(r)

        def subThousand_C(n):
            """Convert a cardinal to words for numbers less than a thousand."""
            if n <= 19:
                return ones_C[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                return tens_C[q] + ("-" + subThousand_C(r) if r else "")
            else:
                q, r = divmod(n, 100)
                return (ones_C[q] +
                        " hundred" +
                        (" and " + subThousand_C(r) if r else ""))

        def thousandUp(n):
            """Return cardinal greater than a thousand in words."""
            list1 = []
            for i, z in enumerate(splitByThousands(n)):
                if i and z:
                    list1.insert(0, subThousand_C(z) + " " + large[i])
                elif z:
                        list1.insert(0, subThousand_C(z))
            return ", ".join(list1)

        def splitByThousands(n):
            "Return reversed digits of n in groups of 3."""
            res = []
            while n:
                n, r = divmod(n, 1000)
                res.append(r)
            return res

        n_clean = ''
        for i in range(len(n)):
            if i == 0:
                if n[i].isdigit() or n[i] == '-':
                    n_clean += n[i]
            else:
                if n[i].isdigit():
                    n_clean += n[i]
        num = int(n_clean)
        if num == 0:
            return "Zero"
        else:
            w = thousandUp(num)
            if len(n_clean) < 4:
                pen = subThousand_O(num)
            elif n_clean[-3] == '0':
                if n_clean[-2] == '0' and n_clean[-1] != '0':
                    ind = w.rfind(" ")
                    end_num = int(n_clean[-1])
                    f = w[:ind - 1] + " and " + subThousand_O(end_num)
                elif n_clean[-2:] != '00':
                    ind = w.rfind(" ")
                    end_num = int(n_clean[-2:])
                    f = w[:ind - 1] + " and " + subThousand_O(end_num)
                elif n_clean[-3:] == '000':
                    ind = w.rfind(" ")
                    f = w[:ind - 1]
                else:
                    f = w
                while not f[-1].isalpha():
                    f = f[:-1]
                pen = f
            elif n_clean[-3] != '0':
                if n_clean[-2] == '0' and n_clean[-1] != '0':
                    ind = w.rfind(" ")
                    end_num = int(n_clean[-1])
                    f = w[:ind] + " " + subThousand_O(end_num)
                elif n_clean[-2:] != '00':
                    ind = w.rfind(" ")
                    end_num = int(n_clean[-2:])
                    f = w[:ind] + " " + subThousand_O(end_num)
                else:
                    f = w
                while not f[-1].isalpha():
                    f = f[:-1]
                pen = f
            if ' , ' in pen:
                pen2 = re.sub('(?<= ), ', '', pen)
                pen2 = re.sub('  ', ' ', pen2)
                pen = pen2
            if pen[-3:] == 'ion' or pen[-3:] == 'and':
                return pen + "th"
            else:
                return pen

    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return n

scurr_dict = {'$': 'dollar', 'Y': 'yen', '€': 'euro',
              '£': 'pound', 'HK$': 'Hong Kong Dollar',
              'US$': 'U S Dollar'}
scurr_dict_pl = {'$': 'dollars', 'Y': 'yen', '€': 'euros',
                 '£': 'pounds', 'HK$': 'Hong Kong Dollars',
                 'US$': 'U S Dollars'}
end_dict = {'k': 'thousand', 'm': 'million', 'b': 'billion'}
ecurr_dict = {
                     "US$": "U S Dollar",
                     "AFN": "Afghanistan Afghani",
                     "ALL": "Albanian Lek",
                     "DZD": "Algerian Dinar",
                     "AOA": "Angolan Kwanza",
                     "ARS": "Argentine Peso",
                     "AMD": "Armenian Dram",
                     "AWG": "Aruban Florin",
                     "AUD": "Australian Dollar",
                     "AZN": "Azerbaijan New Manat",
                     "BSD": "Bahamian Dollar",
                     "BHD": "Bahraini Dinar",
                     "BDT": "Bangladeshi Taka",
                     "BBD": "Barbados Dollar",
                     "BYR": "Belarusian Ruble",
                     "BZD": "Belize Dollar",
                     "BMD": "Bermudian Dollar",
                     "BTN": "Bhutan Ngultrum",
                     "BOB": "Bolivian Boliviano",
                     "BAM": "Bosnian Mark",
                     "BWP": "Botswana Pula",
                     "BRL": "Brazilian Real",
                     "BND": "Brunei Dollar",
                     "BGN": "Bulgarian Lev",
                     "BIF": "Burundi Franc",
                     "XOF": "CFA Franc BCEAO",
                     "XAF": "CFA Franc BEAC",
                     "XPF": "CFP Franc",
                     "KHR": "Cambodian Riel",
                     "CAD": "Canadian Dollar",
                     "CHF": "Swiss Franc",
                     "CVE": "Cape Verde Escudo",
                     "KYD": "Cayman Islands Dollar",
                     "CLP": "Chilean Peso",
                     "CNY": "Chinese Yuan",
                     "COP": "Colombian Peso",
                     "KMF": "Comoros Franc",
                     "CDF": "Congolese Franc",
                     "CRC": "Costa Rican Colon",
                     "HRK": "Croatian Kuna",
                     "CUC": "Cuban Convertible Peso",
                     "CUP": "Cuban Peso",
                     "CYP": "Cyprus Pound",
                     "CZK": "Czech Koruna",
                     "DKK": "Danish Krone",
                     "DJF": "Djibouti Franc",
                     "DOP": "Dominican Republic Peso",
                     "XCD": "East Caribbean Dollar",
                     "EGP": "Egyptian Pound",
                     "SVC": "El Salvador Colon",
                     "EEK": "Estonian Kroon",
                     "ETB": "Ethiopian Birr",
                     "EUR": "Euro",
                     "FKP": "Falkland Islands Pound",
                     "FJD": "Fiji Dollar",
                     "GMD": "Gambian Dalasi",
                     "GEL": "Georgian Lari",
                     "GHS": "Ghanaian New Cedi",
                     "GIP": "Gibraltar Pound",
                     "XAU": "Gold (oz)",
                     "GTQ": "Guatemalan Quetzal",
                     "GNF": "Guinea Franc",
                     "GYD": "Guyanese Dollar",
                     "HTG": "Haitian Gourde",
                     "HNL": "Honduran Lempira",
                     "HKD": "Hong Kong Dollar",
                     "HUF": "Hungarian Forint",
                     "ISK": "Iceland Krona",
                     "INR": "Indian Rupee",
                     "IDR": "Indonesian Rupiah",
                     "IRR": "Iranian Rial",
                     "IQD": "Iraqi Dinar",
                     "ILS": "Israeli New Shekel",
                     "JMD": "Jamaican Dollar",
                     "JPY": "Japanese Yen",
                     "JOD": "Jordanian Dinar",
                     "KZT": "Kazakhstan Tenge",
                     "KES": "Kenyan Shilling",
                     "KWD": "Kuwaiti Dinar",
                     "KGS": "Kyrgyzstanian Som",
                     "LAK": "Lao Kip",
                     "LVL": "Latvian Lats",
                     "LBP": "Lebanese Pound",
                     "LSL": "Lesotho Loti",
                     "LRD": "Liberian Dollar",
                     "LYD": "Libyan Dinar",
                     "LTL": "Lithuanian Litas",
                     "MOP": "Macau Pataca",
                     "MKD": "Macedonian Denar",
                     "MGA": "Malagasy Ariary",
                     "MWK": "Malawi Kwacha",
                     "MYR": "Malaysian Ringgit",
                     "MVR": "Maldive Rufiyaa",
                     "MTL": "Maltese Lira",
                     "MRO": "Mauritanian Ouguiya",
                     "MUR": "Mauritius Rupee",
                     "MXN": "Mexican Peso",
                     "MDL": "Moldovan Leu",
                     "MNT": "Mongolian Tugrik",
                     "MAD": "Moroccan Dirham",
                     "MZN": "Mozambique New Metical",
                     "MMK": "Myanmar Kyat",
                     "ANG": "NL Antillian Guilder",
                     "NAD": "Namibia Dollar",
                     "NPR": "Nepalese Rupee",
                     "NZD": "New Zealand Dollar",
                     "NIO": "Nicaraguan Cordoba Oro",
                     "NGN": "Nigerian Naira",
                     "KPW": "North Korean Won",
                     "NOK": "Norwegian Krone",
                     "OMR": "Omani Rial",
                     "PKR": "Pakistan Rupee",
                     "PAB": "Panamanian Balboa",
                     "PGK": "Papua New Guinea Kina",
                     "PYG": "Paraguay Guarani",
                     "PEN": "Peruvian Nuevo Sol",
                     "PHP": "Philippine Peso ",
                     "PLN": "Polish Zloty",
                     "QAR": "Qatari Rial",
                     "RON": "Romanian New Leu",
                     "RUB": "Russian Rouble",
                     "RWF": "Rwandan Franc",
                     "WST": "Samoan Tala",
                     "STD": "Sao Tome/Principe Dobra",
                     "SAR": "Saudi Riyal",
                     "RSD": "Serbian Dinar",
                     "SCR": "Seychelles Rupee",
                     "SLL": "Sierra Leone Leone",
                     "XAG": "Silver oz",
                     "SGD": "Singapore Dollar",
                     "SKK": "Slovak Koruna",
                     "SIT": "Slovenian Tolar",
                     "SBD": "Solomon Islands Dollar",
                     "SOS": "Somali Shilling",
                     "ZAR": "South African Rand",
                     "KRW": "South-Korean Won",
                     "LKR": "Sri Lanka Rupee",
                     "SHP": "St Helena Pound",
                     "SDG": "Sudanese Pound",
                     "SRD": "Suriname Dollar",
                     "SZL": "Swaziland Lilangeni",
                     "SEK": "Swedish Krona",
                     "CHF": "Swiss Franc",
                     "SYP": "Syrian Pound",
                     "TWD": "Taiwan Dollar",
                     "TZS": "Tanzanian Shilling",
                     "THB": "Thai Baht",
                     "TOP": "Tonga Pa'anga",
                     "TTD": "Trinidad/Tobago Dollar",
                     "TND": "Tunisian Dinar",
                     "TRY": "Turkish New Lira",
                     "TMM": "Turkmenistan Manat",
                     "USD": "US Dollar",
                     "UGX": "Uganda Shilling",
                     "UAH": "Ukraine Hryvnia",
                     "UYU": "Uruguayan Peso",
                     "AED": "United Arab Emir Dirham",
                     "VUV": "Vanuatu Vatu",
                     "VEB": "Venezuelan Bolivar",
                     "VND": "Vietnamese Dong",
                     "YER": "Yemeni Rial",
                     "ZMK": "Zambian Kwacha",
                     "ZWD": "Zimbabwe Dollar",
                     "£": "pound",
                     "$": "dollar",
                     "€": "euro"
                     }
invariant_plural_curr = ['JPY', 'CNY', 'THB', 'ZAR']
irregular_plural_curr = {'SEK': 'Swedish Kronor', 'NOK': 'Norwegian Kroner',
                         'DKK': 'Danish Kroner', 'ISK': 'Iceland Kronur',
                         'BRL': 'Brazilian Reais', 'LTL': 'Litai',
                         'LVL': 'Latu', 'RON': 'Lei'}


def expand_MONEY(dict_tup, text):
    try:
        """Return number and currency of input in words."""
        ind, (n, tag, ntag) = dict_tup
        scurr = ''
        ecurr = ''
        end = ''
        num = ''
        rcontext = gen_frame(dict_tup, text)[3]
        if rcontext in bmoney:
            s1 = expand_MONEY((1, (n, tag, ntag)), ['the', n, 'drop'])
            inds = [ind + 1 for ind in range(len(s1)) if s1[ind] == ' ']
            inds = sorted(inds, reverse=True)
            split = len(s1)
            currency = ''
            possibles = (set(scurr_dict.values()) | set(scurr_dict_pl.values())
                         | set(ecurr_dict.values())
                         | set(irregular_plural_curr.values())
                         | {it + 's' for it in ecurr_dict.values()})
            for i in inds:
                curr_cand = s1[i:]
                if curr_cand in possibles:
                    split = i
                    currency = curr_cand
                else:
                    pass
            return s1[:split] + rcontext + " " + currency

        if len(n) <= 3:
            if not n[0].isdigit():
                if len(n) == 3 and not n[2].isdigit():
                    scurr += n[0]
                    num += n[1]
                    end += n[2]
                else:
                    scurr += n[0]
                    num += n[1:]
            else:
                num += n[0]
                ecurr += n[1:]
        else:
            for i in range(len(n) - 3):
                if not n[i].isdigit() and not n[i] == '.':
                    scurr += n[i]
                else:
                    num = n[i:-3]
                    break
            if n[-3].isalpha():
                ecurr += n[-3:]
            elif n[-1].isalpha():
                num += n[-3:-1]
                end += n[-1]
            else:
                num += n[-3:]

        if '-' in num:
            exp_num = expand_NRANGE(num)
        else:
            exp_num = expand_NUM(num)
        if num == '1' and not end:
            if ecurr:
                currency = ecurr_dict[ecurr]
            elif scurr:
                currency = scurr_dict[scurr]
            else:
                currency = ''
        else:
            if ecurr:
                if ecurr in invariant_plural_curr:
                    currency = ecurr_dict[ecurr]
                elif ecurr in irregular_plural_curr:
                    currency = irregular_plural_curr[ecurr]
                else:
                    currency = ecurr_dict[ecurr] + 's'
            elif scurr:
                currency = scurr_dict_pl[scurr]

            else:
                currency = ''
        if end:
            large = end_dict[end.lower()]
        else:
            large = ''
        items = [exp_num]
        if large:
            items.append(large)
        items.append(currency)

        return ' '.join(items)
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return n


def expand_NDIG(w):
    """Expand digit sequences to a series of words."""
    try:
        numbers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        num_words = ['zero', 'one', 'two', 'three', 'four', 'five',
                     'six', 'seven', 'eight', 'nine']
        str2 = ''
        for n in w:
            if n.isdigit():
                str2 += num_words[numbers.index(n)]
                str2 += " "
            elif n == '.':
                str2 += 'dot'
                str2 += ' '
            else:
                str2 += ''
        return str2[:-1]
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w


def expand_NTEL(w):
    """Expand telephone numbers."""
    try:
        str2 = ''
        if w[0] == '+':
            str2 += 'plus '
            str2 += expand_NDIG(w[1:])
        else:
            str2 += expand_NDIG(w)
        return str2
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w


numbers = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
num_words = ['zero', 'one', 'two', 'three', 'four', 'five',
             'six', 'seven', 'eight', 'nine']
numbers1 = ['00', '01', '02', '03', '04', '05',
            '06', '07', '08', '09', '10', '11', '12']
num_words1 = ['zero', 'one', 'two', 'three', 'four', 'five',
              'six', 'seven', 'eight', 'nine', 'ten', 'eleven', 'twelve']
numbers2 = ['13', '14', '15', '16', '17', '18', '19', '20', '21', '22', '23']
num_words2 = ['one', 'two', 'three', 'four', 'five',
              'six', 'seven', 'eight', 'nine', 'ten', 'eleven']


def expand_NTIME(w):
    """Expand a time to words."""
    try:
        str2 = ''
        m = time_pattern.match(w)
        if w in ['0:00' or '00:00']:
            return 'midnight'
        if m.group(1) in numbers:
            str2 += num_words[numbers.index(m.group(1))]
            str2 += ' '
        elif m.group(1) in numbers1:
            str2 += num_words1[numbers1.index(m.group(1))]
            str2 += ' '
        elif m.group(1) in numbers2:
            str2 += num_words2[numbers2.index(m.group(1))]
            str2 += ' '
        if m.group(3) == '00':
            str2 += ''
        elif int(m.group(3)) < 10:
            str2 += expand_NDIG(m.group(3))
        else:
            str2 += expand_NUM(m.group(3))
            str2 += ' '
        return str2
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w


def expand_NYER(w):
    """Expand a year to words."""
    try:
        if w[-1] == 's':
            if len(w) > 1 and w[-2:] == "'s":
                str = ''
                str += expand_NYER(w[:-2])
                str += "'s"
            else:
                str = ''
                if expand_NYER(w[:-1])[-1] == 'y':
                    str += expand_NYER(w[:-1])[:-1]
                    str += 'ies'
                else:
                    str += expand_NYER(w[:-1])
                    str += 's'
            return str

        num_decades = ['00s', '10s', '20s', '30s', '40s', '50s',
                       '60s', '70s', '80s', '90s']
        decades = ['hundreds', 'tens', 'twenties', 'thirties', 'forties',
                   'fifties', 'sixties', 'seventies', 'eighties', 'nineties']
        if w[-3:] in num_decades:
            return expand_NUM(w[:2]) + " " + decades[num_decades.index(w[-3:])]
        elif w[1:3] == '00':
            return expand_NUM(w)
        elif w[2:4] == '00':
            for i in range(len(w)):
                a = w[:2]
                return expand_NUM(a) + " " + "hundred"
        else:
            if len(w) == 4:
                for i in range(len(w)):
                    a = w[:2]
                    b = w[2:]
                    return expand_NUM(a) + " " + expand_NUM(b)
            else:
                return expand_NUM(w)

    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w


def expand_NDATE(w, variety="BrE"):
    """Expand a date to words."""
    try:
        numbers = ['01', '1', '02', '2', '03', '3', '04', '4', '05', '5', '06',
                   '6', '07', '7', '08', '8', '09', '9', '10', '11', '12']
        months = ['January', 'January', 'February', 'February', 'March',
                  'March', 'April', 'April', 'May', 'May', 'June', 'June',
                  'July', 'July', 'August', 'August', 'September', 'September',
                  'October', 'November', 'December']
        m = date_pattern.match(w)
        str2 = ''
        if variety in ["BrE", "BrEng"]:
            if m.group(5):
                str2 += 'the ' + (expand_ordinal(m.group(1)) + ' of '
                        + months[numbers.index(m.group(3))]
                        + " " + expand_NYER(m.group(5)))
            else:
                str2 += 'the ' + (expand_ordinal(m.group(1)) + ' of '
                        + months[numbers.index(m.group(3))])
        elif variety in ["AmE", "AmEng"]:
            if m.group(5):
                str2 += 'the ' + (expand_ordinal(m.group(3)) + ' of '
                        + months[numbers.index(m.group(1))]
                        + " " + expand_NYER(m.group(5)))
            else:
                str2 += 'the ' + (expand_ordinal(m.group(3)) + ' of '
                        + months[numbers.index(m.group(1))])
        else:
            if m.group(5):
                str2 += 'the ' + (expand_ordinal(m.group(1)) + ' of '
                        + months[numbers.index(m.group(3))]
                        + " " + expand_NYER(m.group(5)))
            else:
                str2 += 'the ' + (expand_ordinal(m.group(1)) + ' of '
                        + months[numbers.index(m.group(3))])
        return str2
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w



def expand_PRCT(w):
    """Expand a percentage to words."""
    try:
        if '-' in w or '–' in w:
            return expand_NRANGE(w[:-1]) + " percent"
        elif '.' in w:
            m = percent_pattern2.match(w)
            if m:
                a = m.group(1)
                b = m.group(3)
                return expand_NUM(a) + " point " + expand_NDIG(b) + " percent"
            else:
                return w
        else:
            m = percent_pattern1.match(w)
            if m:
                a = m.group(1)
                return expand_NUM(a) + " percent"
            elif whole_and_fract_pattern.match(w[:-1]):
                return expand_NUM(w[:-1]) + " percent"
            elif "/" in w:
                if w[:-1] == '1/2':
                    return expand_fraction(w[:-1]) + " a percent"
                else:
                    return expand_fraction(w[:-1]) + " of a percent"
            else:
                return w
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w


def expand_NSCI(w):
    """Expand scientific numbers (coordinates, feet and inches) to words."""
    try:
        if coord_pattern.match(w):
            m = coord_pattern.match(w)
            exp = ''
            if m.group(1):
                exp += expand_NUM(m.group(1)[:-1]) + " degrees"
                if m.group(2):
                    exp += ", "
            if m.group(2):
                exp += expand_NUM(m.group(2)[:-1]) + " minutes"
                if m.group(3):
                    exp += ", "
            if m.group(3):
                if m.group(3).isdigit():
                    exp += expand_NUM(m.group(3)) + " seconds"
                else:
                    exp += expand_NUM(m.group(3)[:-1]) + " seconds"
                if m.group(4):
                    exp += ", "
            if m.group(4):
                if m.group(4) == "N":
                    exp += "North"
                if m.group(4) == "S":
                    exp += "South"
                if m.group(4) == "W":
                    exp += "West"
                if m.group(4) == "E":
                    exp += "East"
            return exp
        elif feet_pattern.match(w):
            m = feet_pattern.match(w)
            exp = ''
            if m.group(1):
                if m.group(2):
                    if m.group(2).startswith("."):
                        exp += expand_NUM(m.group(1)[:-1]) + " "
                        exp += expand_NUM(m.group(2)) + " feet"
                    else:
                        exp += expand_NUM(m.group(1)[:-1]) + " foot "
                        if m.group(2).isdigit():
                            exp += expand_NUM(m.group(2)) + " inches"
                        else:
                            exp += expand_NUM(m.group(2)[:-1]) + " inches"
                else:
                    exp += expand_NUM(m.group(1)) + " foot"
            return exp
    except (KeyboardInterrupt, SystemExit):
        raise
    except:
        return w

percent_pattern1 = re.compile('''
([0-9]+)
(%)
$
''', re.VERBOSE)

percent_pattern2 = re.compile('''
([0-9]+)
([\.]?)
([0-9]+?)
(%)
$
''', re.VERBOSE)

time_pattern = re.compile('''
([0-9]{1,2})
([\.|:])
([0-9]{2})
$
''', re.VERBOSE)

date_pattern = re.compile('''
([0-9]{1,2})
([\.|-|./])
([0-9]{1,2})
([\.|-|./]?)
([0-9]{0,4})
$
''', re.VERBOSE)

range_pattern = re.compile('''
([0-9]+                 # 1 or more digits
\.?                     # optional '.'
[0-9]*)                 # 0 or more digits
(/|-|–)                   # hyphen or slash
([0-9]+                 # same pattern as lines 1-3
\.?
[0-9]*)
$
''', re.VERBOSE)

dec2_pattern = re.compile('''
([0-9]+)
[\.|:]
([0-9]+
\.
[0-9]+)
$
''', re.VERBOSE)

whole_and_fract_pattern = re.compile('''
([0-9]+)
[-]
([0-9]+
/[0-9]+)
$
''', re.VERBOSE)

coord_pattern = re.compile('''
([0-9]+
\.?
[0-9]*
°)
([0-9]+
\.?
[0-9]*
[\'|\’|′])?
([0-9]+
\.?
[0-9]*
["|″]?)?
([N|S|E|W])?
$
''', re.VERBOSE)

feet_pattern = re.compile('''
([0-9]+
[\'|\’|′])
(\.?
[0-9]+
["|″]?)?
$
''', re.VERBOSE)

bmoney = [
        "thousand", "million", "billion", "trillion", "quadrillion",
        "quintillion", "sextillion", "septillion", "octillion",
        "nonillion", "decillion", "undecillion", "duodecillion",
        "tredecillion", "quattuordecillion", "sexdecillion",
        "septendecillion", "octodecillion", "novemdecillion",
        "vigintillion"
        ]

months = ['Jan', 'Jan.', 'January', 'Feb', 'Feb.', 'February', 'Mar',
          'Mar.', 'March', 'Apr', 'Apr.', 'April', 'May', 'Jun', 'Jun.',
          'June', 'Jul', 'Jul.', 'July', 'Aug', 'Aug.', 'August',
          'Sept', 'Sept.', 'September', 'Oct', 'Oct.', 'October',
          'Nov', 'Nov.', 'November', 'Dec', 'Dec.', 'December']



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import pickle
from io import open

from nltk.corpus import names

from normalise.detect import create_NSW_dict, mod_path
from normalise.tagger import tagify, is_digbased
from normalise.splitter import split, retagify
from normalise.class_ALPHA import run_clfALPHA
from normalise.class_NUMB import run_clfNUMB, gen_frame
from normalise.tag_MISC import tag_MISC
from normalise.expand_all import expand_all
from normalise.expand_NUMB import bmoney


with open('{}/data/wordlist.pickle'.format(mod_path), mode='rb') as file:
    wordlist = pickle.load(file)

with open('{}/data/names.pickle'.format(mod_path), mode='rb') as file:
    names_lower = pickle.load(file)


def list_NSWs(text, verbose=True, variety='BrE', user_abbrevs={}):
    if verbose:
        print("\nCREATING NSW DICTIONARY")
        print("-----------------------\n")

    NSWs = create_NSW_dict(text, verbose=verbose)
    if verbose:
        print("{} NSWs found\n".format(len(NSWs)))
        print("TAGGING NSWs")
        print("------------\n")
    tagged = tagify(NSWs, verbose=verbose)
    ALPHA_dict = {}
    NUMB_dict = {}
    MISC_dict = {}
    SPLT_dict = {}
    for item in tagged.items():
        tag = item[1][1]
        if tag == 'ALPHA':
            ALPHA_dict.update((item,))
        elif tag == 'NUMB':
            NUMB_dict.update((item,))
        elif tag == 'MISC':
            MISC_dict.update((item,))
        elif tag == 'SPLT':
            SPLT_dict.update((item,))
    if verbose:
        print("SPLITTING NSWs")
        print("--------------\n")
    splitted = split(SPLT_dict, verbose=verbose)
    if verbose:
        print("RETAGGING SPLIT NSWs")
        print("--------------------\n")
    retagged = retagify(splitted, verbose=verbose)
    for item in retagged.items():
        tag = item[1][1]
        if tag == 'SPLT-ALPHA':
            ALPHA_dict.update((item,))
        elif tag == 'SPLT-NUMB':
            NUMB_dict.update((item,))
        elif tag == 'SPLT-MISC':
            MISC_dict.update((item,))
    if verbose:
        print("CLASSIFYING ALPHABETIC NSWs")
        print("---------------------------\n")
    tagged_ALPHA = run_clfALPHA(ALPHA_dict, text, verbose=verbose, user_abbrevs=user_abbrevs)
    if verbose:
        print("CLASSIFYING NUMERIC NSWs")
        print("------------------------\n")
    tagged_NUMB = run_clfNUMB(NUMB_dict, text, verbose=verbose)
    if verbose:
        print("CLASSIFYING MISCELLANEOUS NSWs")
        print("------------------------------\n")
    tagged_MISC = tag_MISC(MISC_dict, verbose=verbose)
    if verbose:
        print("EXPANDING ALPHABETIC NSWs")
        print("-------------------------\n")
    expanded_ALPHA = expand_all(tagged_ALPHA, text, verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
    if verbose:
        print("EXPANDING NUMERIC NSWs")
        print("----------------------\n")
    expanded_NUMB = expand_all(tagged_NUMB, text, verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
    if verbose:
        print("EXPANDING MISCELLANEOUS NSWs")
        print("----------------------------\n")
    expanded_MISC = expand_all(tagged_MISC, text, verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
    return expanded_ALPHA, expanded_NUMB, expanded_MISC


def tokenize_basic(text):
    guess = [d for w in text.split(' ') for d in w.split('\n')]
    out = []
    for i in range(len(guess) - 1):
        if not guess[i]:
            pass
        elif guess[i].isalpha():
            out.append(guess[i])
        elif guess[i][0] in ['(', '[', '{']:
            if guess[i][1] in [')', ']', '}']:
                out.extend([guess[i][0], guess[i][1:-1], guess[i][-1]])
            else:
                out.extend([guess[i][0], guess[i][1:]])
        elif guess[i][-1] in [')', ']', '}']:
                out.extend([guess[i][:-1], guess[i][-1]])
        elif guess[i][-1] in ['!', '?'] and guess[i][:-1].isalpha():
            out.extend([guess[i][:-1], guess[i][-1]])
        elif guess[i][-1] == '.' and guess[i][:-1].isalpha():
            following = guess[i + 1]
            if following.istitle() and following.lower() in wordlist:
                if following.lower() in names_lower:
                    if guess[i][:-1] in wordlist:
                        out.extend([guess[i][:-1], '.'])
                    else:
                        out.append(guess[i])
                else:
                    out.extend([guess[i][:-1], '.'])
            elif guess[i][-1] == '.' and is_digbased(guess[i][:-1]):
                out.extend([guess[i][:-1], '.'])
            else:
                out.append(guess[i])
        elif guess[i].endswith((',', ':', ';')):
            out.extend([guess[i][:-1], guess[i][-1]])
        else:
            out.append(guess[i])
    if not guess[-1]:
        pass
    elif guess[-1].isalpha():
        out.append(guess[-1])
    elif guess[-1][-1] in ['!', '?'] and guess[-1][:-1].isalpha():
        out.extend([guess[-1][:-1], guess[-1][-1]])
    elif guess[-1][-1] == '.' and guess[-1][:-1] in wordlist:
        out.extend([guess[-1][:-1], '.'])
    elif guess[-1][-1] == '.' and is_digbased(guess[-1][:-1]):
        out.extend([guess[-1][:-1], '.'])
    elif guess[-1].endswith((',', ':', ';')):
        out.extend([guess[-1][:-1], guess[-1][-1]])
    else:
        out.append(guess[-1])
    return out


def normalise(text, tokenizer=tokenize_basic, verbose=True, variety='BrE', user_abbrevs={}):
    if type(text) == str:
        if tokenizer == tokenize_basic and verbose:
            print("NOTE: using basic tokenizer.\n"
                  "For better results, input tokenized text,"
                  " or use a custom tokenizer")
            return insert(tokenizer(text), verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
        else:
            return insert(tokenizer(text), verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
    else:
        return insert(text, verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)


def insert(text, verbose=True, variety='BrE', user_abbrevs={}):
    (expanded_ALPHA,
    expanded_NUMB,
    expanded_MISC) = list_NSWs(text, verbose=verbose, variety=variety, user_abbrevs=user_abbrevs)
    out = text[:]
    split_dict = {}
    for item in (expanded_ALPHA, expanded_NUMB, expanded_MISC):
        for nsw in item.items():
            if isinstance(nsw[0], int):
                out[nsw[0]] = nsw[1][3]
                if nsw[1][2] == 'MONEY' and gen_frame(nsw, text)[3] in bmoney:
                    out[nsw[0] + 1] = ''
            else:
                rind = int(nsw[0])
                if rind in split_dict:
                    split_dict[rind][100 * (nsw[0] - rind)] = nsw[1][3]
                else:
                    split_dict[rind] = {(100 * (nsw[0] - rind)): nsw[1][3]}
                if out[rind] == text[rind]:
                    out[rind] = nsw[1][3]
                else:
                    final = ''
                    for it in sorted(split_dict[rind]):
                        final += ' '
                        final += split_dict[rind][it]
                    final = final[1:]
                    out[rind] = final
    return out


def rejoin(tokenized_text):
    out = ''
    for word in tokenized_text:
        if word:
            out += word
            out += ' '
    return out



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import pickle
from collections import defaultdict

from normalise.detect import mod_path

abbrev_dict = {
               "abbrev.": "abbreviation",
               "abr.": "abridged",
               "abridg.": "abridged",
               "absol.": "absolute",
               "abst.": "abstract",
               "abstr.": "abstract",
               "accomm.": "accommodation",
               "accompl.": "accomplished",
               "acct.": "account",
               "accts.": "accounts",
               "addr.": "address",
               "adj.": "adjective",
               "adjs.": "adjectives",
               "adm.": "admiral",
               "adv.": "adverb",
               "advb.": "adverb",
               "advt.": "advertisement",
               "advts.": "advertisements",
               "aff.": "affairs",
               "afr.": "Africa",
               "agst.": "against",
               "alg.": "algebra",
               "alt.": "alteration",
               "ann.": "annual",
               "anniv.": "anniversary",
               "anon.": "anonymous",
               "answ.": "answer",
               "apol.": "apology",
               "apr.": "April",
               "arch.": "architecture",
               "argt.": "argument",
               "arith.": "arithmetic",
               "artific.": "artificial",
               "assemb.": "assembly",
               "assoc.": "association",
               "astr.": "astronomy",
               "attrib.": "attribute",
               "aug.": "August",
               "bef.": "before",
               "belg.": "Belgian",
               "betw.": "between",
               "biog.": "biography",
               "bk.": "book",
               "bks.": "books",
               "braz.": "Brazilian",
               "c.": "century",
               "c": "circa",
               "cal.": "calendar",
               "calc.": "calculus",
               "calif.": "California",
               "camb.": "Cambridge",
               "capt.": "captain",
               "cath.": "Catholic",
               "cent.": "century",
               "cert.": "certificate",
               "certif.": "certificate",
               "char.": "character",
               "circ.": "circle",
               "cl.": "clause",
               "co.": "company",
               "col.": "colonel",
               "coll.": "college",
               "colloq.": "colloquial",
               "comb.": "combination",
               "combs.": "combinations",
               "commerc.": "commercial",
               "commonw.": "commonwealth",
               "conc.": "concerning",
               "concl.": "conclusion",
               "conf.": "conference",
               "congr.": "congress",
               "conj.": "conjunction",
               "conn.": "Connecticut",
               "cons.": "consonant",
               "constr.": "construction",
               "contemp.": "contemporary",
               "cont.": "continuation",
               "contrib.": "contribution",
               "conv.": "convention",
               "counc.": "council",
               "cmpd.": "compound",
               "crim.": "criminal",
               "crt.": "court",
               "crts.": "courts",
               "dec.": "December",
               "dept.": "department",
               "devel.": "development",
               "dict.": "dictionary",
               "dist.": "district",
               "distrib.": "distribution",
               "dk.": "Duke",
               "docs.": "documents",
               "doc.": "document",
               "dr.": "Doctor",
               "e.": "east",
               "eds.": "editions",
               "ed.": "edition",
               "edw.": "Edward",
               "emb.": "embassy",
               "encycl.": "encyclopaedia",
               "enq.": "enquiry",
               "equip.": "equipment",
               "esp.": "especially",
               "eval.": "evaluation",
               "evid.": "evidence",
               "exam.": "examination",
               "exc.": "except",
               "exch.": "exchange",
               "exerc.": "exercise",
               "f.": "feminine",
               "feb.": "February",
               "fem.": "feminine",
               "figs.": "figures",
               "fig.": "figure",
               "freq.": "frequently",
               "gd.": "good",
               "gen.": "general",
               "geog.": "geography",
               "gloss.": "glossary",
               "govt.": "government",
               "gr.": "grammar",
               "gt.": "great",
               "handbk.": "handbook",
               "hist.": "historical",
               "hosp.": "hospital",
               "husb.": "husband",
               "inc.": "incorporated",
               "ind.": "industry",
               "indef.": "indefinite",
               "inq.": "inquiry",
               "intro.": "introduction",
               "inv.": "inventory",
               "investig.": "investigation",
               "investm.": "investment",
               "irel.": "Ireland",
               "irreg.": "irregular",
               "jan.": "January",
               "jr.": "junior",
               "jrnl.": "journal",
               "jrnls.": "journals",
               "jul.": "July",
               "jun.": "June",
               "k.": "king",
               "kingd.": "kingdom",
               "knowl.": "knowledge",
               "lab.": "laboratory",
               "lang.": "language",
               "langs.": "languages",
               "lat.": "Latin",
               "ld.": "Lord",
               "lds.": "Lords",
               "lect.": "lecture",
               "let.": "letter",
               "letts.": "letters",
               "libr.": "library",
               "ling.": "linguistic",
               "lt.": "Lieutenant",
               "m.": "masculine",
               "mag.": "magazine",
               "man.": "manual",
               "mar.": "March",
               "masc.": "masculine",
               "meas.": "measure",
               "med.": "medieval",
               "mem.": "memoir",
               "merch.": "merchandise",
               "meth.": "method",
               "mil.": "military",
               "milit.": "military",
               "mispr.": "misprinted",
               "mod.": "modern",
               "MS.": "manuscript",
               "mtg.": "meeting",
               "mts.": "mountains",
               "mus.": "museum",
               "myst.": "mystery",
               "myth.": "mythology",
               "narr.": "narrative",
               "nat.": "natural",
               "naut.": "nautical",
               "nav.": "naval",
               "navig.": "navigation",
               "neurol.": "neurology",
               "no.": "number",
               "norweg.": "Norwegian",
               "nov.": "November",
               "num.": "Numbers",
               "N.Z.": "New Zealand",
               "observ.": "observation",
               "occas.": "occasion",
               "occurr.": "occurrence",
               "oct.": "October",
               "O.E.D.": "Oxford English Dictionary",
               "offic.": "official",
               "org.": "organic",
               "orig.": "original",
               "oxf.": "Oxford",
               "p.": "page",
               "pass.": "passive",
               "perf.": "perfect",
               "pers.": "personal",
               "photog.": "photography",
               "photogr.": "photography",
               "phr.": "phrase",
               "pict.": "picture",
               "pl.": "plural",
               "plur.": "plural",
               "pop.": "popular",
               "ppl.": "people",
               "pr.": "present",
               "prec.": "preceding",
               "pred.": "prediction",
               "pref.": "prefix",
               "prob.": "probably",
               "probl.": "problem",
               "prod.": "product",
               "pron.": "pronoun",
               "pronunc.": "pronunciation",
               "prop.": "properly",
               "prov.": "proverb",
               "provid.": "providence",
               "provinc.": "provincial",
               "psych.": "psychology",
               "pt.": "part",
               "q.": "quarterly",
               "Q.": "queen",
               "quot.": "quotation",
               "quots.": "quotations",
               "R.C.": "Roman Catholic",
               "rd.": "road",
               "ref.": "reference",
               "regist.": "register",
               "rel.": "relative",
               "rept.": "report",
               "repub.": "republic",
               "ret.": "return",
               "revol.": "revolution",
               "rich.": "Richard",
               "R.N.": "Royal Navy",
               "rom.": "Roman",
               "s.": "south",
               "sat.": "Saturday",
               "sax.": "Saxon",
               "sch.": "school",
               "scot.": "Scottish",
               "scotl.": "Scotland",
               "sen.": "Senator",
               "sept.": "September",
               "sess.": "session",
               "settlem.": "settlement",
               "sev.": "several",
               "soc.": "society",
               "sp.": "spelling",
               "span.": "Spanish",
               "spec.": "specimen",
               "std.": "standard",
               "stand.": "standard",
               "str.": "strong",
               "subj.": "subject",
               "subord.": "subordinate",
               "syll.": "syllable",
               "symmetr.": "symmetrical",
               "syst.": "system",
               "techn.": "technical",
               "telecomm.": "telecommunications",
               "test.": "testament",
               "textbk.": "textbook",
               "thes.": "thesaurus",
               "trag.": "tragedy",
               "transf.": "transfer",
               "transl.": "translation",
               "treatm.": "treatment",
               "trig.": "trigonometry",
               "trop.": "tropical",
               "typogr.": "typography",
               "U.K.": "United Kingdom",
               "univ.": "university",
               "unkn.": "unknown",
               "unoffic.": "unofficial",
               "vac.": "vacation",
               "var.": "variant",
               "vb.": "verb",
               "vbl.": "verbal",
               "vbs.": "verbs",
               "veg.": "vegetable",
               "vic.": "Victoria",
               "vict.": "Victoria",
               "vocab.": "vocabulary",
               "vol.": "volume",
               "vols.": "volumes",
               "w.": "west",
               "wd.": "word",
               "wis.": "Wisconsin",
               "wkly.": "weekly",
               "wks.": "works",
               "yearbk.": "yearbook",
               "yng.": "young",
               "yr.": "year",
               "yrs.": "years"
               }


ambig_abbrevs = {
                 "acad.": ["academia", "academy", "academic"],
                 "acc.": ["according", "account"],
                 "admin.": ["administration", "administrative"],
                 "adv.": ["advanced", "adventure"],
                 "agric.": ["agriculture", "agricultural"],
                 "amer.": ["America", "American"],
                 "art.": ["article", "artifact"],
                 "biochem.": ["biochemistry", "biochemical"],
                 "chem.": ["chemistry", "chemical"],
                 "chron.": ["chronology", "chronicle"],
                 "comm.": ["commerce", "committee"],
                 "corresp.": ["corresponding", "correspondence"],
                 "crit.": ["critical", "criticism"],
                 "ct.": ["count", "court"],
                 "def.": ["definite", "definition"],
                 "dep.": ["department", "deputy"],
                 "descr.": ["description", "descriptive"],
                 "e.": ["east", "eastern"],
                 "ecol.": ["ecology", "ecological"],
                 "econ.": ["economics", "economical"],
                 "educ.": ["education", "educational"],
                 "electr.": ["electricity", "electrical"],
                 "fam.": ["familiar", "family"],
                 "eng.": ["England", "English"],
                 "fr.": ["French", "France"],
                 "geogr.": ["geography", "geographical"],
                 "geol.": ["geology", "geological"],
                 "geom.": ["geometry", "geometrical"],
                 "ger.": ["German", "Germany"],
                 "gov.": ["government", "Governor"],
                 "hist.": ["history", "historical"],
                 "illustr.": ["illustrated", "illustration"],
                 "ind.": ["India", "Indian"],
                 "industr.": ["industrial", "industry"],
                 "inst.": ["institute", "institution"],
                 "intell.": ["intelligence", "intelligent"],
                 "introd.": ["introduction", "introductory"],
                 "ital.": ["Italian", "Italy"],
                 "jap.": ["Japanese", "Japan"],
                 "lit.": ["literal", "literally"],
                 "lit.": ["literary", "literature"],
                 "mach.": ["machinery", "machine"],
                 "math.": ["mathematics", "mathematical"],
                 "mech.": ["mechanics", "mechanical"],
                 "med.": ["medicine", "medical"],
                 "mex.": ["Mexican", "Mexico"],
                 "misc.": ["miscellaneous", "miscellany"],
                 "mt.": ["mount", "mountain"],
                 "mus.": ["music", "musical"],
                 "n.": ["north", "northern"],
                 "n.e.": ["north-east", "north-eastern"],
                 "n.w.": ["north-west", "north-western"],
                 "obj.": ["object", "objective"],
                 "occup.": ["occupation", "occupational"],
                 "off.": ["official", "office"],
                 "opp.": ["opposed", "opposite"],
                 "orig.": ["origin", "original"],
                 "parl.": ["parliament", "parliamentary"],
                 "perf.": ["perfect", "perfection"],
                 "phil.": ["philosophy", "philosophical"],
                 "phys.": ["physiology", "physics"],
                 "physiol.": ["physiology", "physiological"],
                 "pol.": ["politics", "political"],
                 "polit.": ["politics", "political"],
                 "port.": ["Portuguese", "Portugal"],
                 "poss.": ["possible", "possibly"],
                 "pract.": ["practical", "practice"],
                 "pres.": ["President", "present"],
                 "princ.": ["principle", "principality"],
                 "probab.": ["probability", "probably"],
                 "prop.": ["property", "proper"],
                 "psychol.": ["psychology", "psychological"],
                 "publ.": ["public", "publication"],
                 "rec.": ["record", "recording"],
                 "reg.": ["regular", "register"],
                 "rep.": ["report", "representative"],
                 "repr.": ["representative", "representing"],
                 "resid.": ["residence", "residential"],
                 "rev.": ["Reverend", "revised", "review"],
                 "russ.": ["Russia", "Russian"],
                 "s.": ["south", "southern"],
                 "scand.": ["Scandinavia", "Scandinavian"],
                 "sci.": ["science", "scientific"],
                 "s.e.": ["south-east", "south-eastern"],
                 "sociol.": ["sociology", "sociological"],
                 "st.": ["Saint", "street"],
                 "struct.": ["structure", "structural"],
                 "suppl.": ["supplement", "supplementary"],
                 "surg.": ["surgery", "surgical"],
                 "surv.": ["survey", "surveying"],
                 "s.w.": ["south-west", "south-western"],
                 "technol.": ["technology", "technological"],
                 "tel.": ["telegraph", "telephone"],
                 "teleph.": ["telephone", "telephony", "telephonic"],
                 "theol.": ["theology", "theological"],
                 "tr.": ["translation", "translating"],
                 "trad.": ["tradition", "traditional"],
                 "transl.": ["translation", "translating"],
                 "treas.": ["treasurer", "treasury"],
                 "w.": ["west", "western"],
                 "wk.": ["week", "work"]
                 }


states = {
          "Ala.": "Alabama",
          "Alaska": "Alaska",
          "Ariz.": "Arizona",
          "Ark.": "Arkansas",
          "Calif.": "California",
          "Colo.": "Colorado",
          "Conn.": "Connecticut",
          "Del.": "Delaware",
          "D.C.": "Dist. of Columbia",
          "Fla.": "Florida",
          "Ga.": "Georgia",
          "Ill.": "Illinois",
          "Ind.": "Indiana",
          "Kans.": "Kansas",
          "Ky.": "Kentucky",
          "La.": "Louisiana",
          "Md.": "Maryland",
          "Mass.": "Massachusetts",
          "Mich.": "Michigan",
          "Minn.": "Minnesota",
          "Miss.": "Mississippi",
          "Mo.": "Missouri",
          "Mont.": "Montana",
          "Nebr.": "Nebraska",
          "Nev.": "Nevada",
          "N.H.": "New Hampshire",
          "N.J.": "New Jersey",
          "N.M.": "New Mexico",
          "N.Y.": "New York",
          "N.C.": "North Carolina",
          "N.D.": "North Dakota",
          "Okla.": "Oklahoma",
          "Ore.": "Oregon",
          "Pa.": "Pennsylvania",
          "P.R.": "Puerto Rico",
          "R.I.": "Rhode Island",
          "S.C.": "South Carolina",
          "S.D.": "South Dakota",
          "Tenn.": "Tennessee",
          "Tex.": "Texas",
          "Vt.": "Vermont",
          "Va.": "Virginia",
          "V.I.": "Virgin Islands",
          "Wash.": "Washington",
          "W.Va.": "West Virginia",
          "Wis.": "Wisconsin",
          "Wyo.": "Wyoming"
          }

titles = {
          "lt.": "Lieutenant",
          "rt.": "right",
          "hon.": "Honourable",
          "supt.": "Superintendent",
          "sr.": "senior"
          }


def build_abbrevs(dictionary):
    """Builds abbreviation dictionary from input in correct format."""
    abbrevs = defaultdict(list)
    for key in dictionary:
        if key.endswith('.'):
            k = key[:-1].lower()
        else:
            k = key.lower()
        if type(dictionary[key]) == list:
            abbrevs[k].extend(dictionary[key])
        elif type(dictionary[key]) == str:
            abbrevs[k].append(dictionary[key])
    return abbrevs


def create_user_abbrevs(dictionary):
    """Stores user and general abbreviations in pickle file."""
    with open('{}/data/abbrev_dict.pickle'.format(mod_path), mode='rb') as f:
        abbrevs = pickle.load(f)
    for key in dictionary:
        if key.endswith('.'):
            k = key[:-1].lower()
        else:
            k = key.lower()
        if type(dictionary[key]) == list:
            for exp in dictionary[key]:
                if exp not in abbrevs[k]:
                    abbrevs[k].append(exp)
        elif type(dictionary[key]) == str:
            if dictionary[key] not in abbrevs[k]:
                abbrevs[k].append(dictionary[key])
    return abbrevs


def add_to_pickled_abbrev(dictionary):
    """Add new dictionary to pickled file of abbreviations."""
    with open('{}/data/abbrev_dict.pickle'.format(mod_path), mode='rb') as f:
        abbrevs = pickle.load(f)
    for key in dictionary:
        if key.endswith('.'):
            k = key[:-1].lower()
        else:
            k = key.lower()
        if type(dictionary[key]) == list:
            for exp in dictionary[key]:
                if exp not in abbrevs[k]:
                    abbrevs[k].append(exp)
        elif type(dictionary[key]) == str:
            if dictionary[key] not in abbrevs[k]:
                abbrevs[k].append(dictionary[key])
    with open('{}/data/abbrev_dict.pickle'.format(mod_path), mode='wb') as f:
        pickle.dump(abbrevs, f, protocol=2)
    print(abbrevs)



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import sys
import re
import pickle
from io import open

import numpy as np
from sklearn.semi_supervised import LabelPropagation as lp

from normalise.detect import mod_path
from normalise.tagger import tagify, ecurr_dict
from normalise.data.timezones import timezone_dict
from normalise.splitter import split, retagify
from normalise.data.measurements import meas_dict, meas_dict_pl


with open('{}/data/NSW_dict.pickle'.format(mod_path), mode='rb') as file:
    NSWs = pickle.load(file)

with open('{}/data/word_tokenized.pickle'.format(mod_path), mode='rb') as file:
    word_tokenized = pickle.load(file)

with open('{}/data/word_tokenized_lowered.pickle'.format(mod_path), mode='rb') as f:
    word_tokenized_lowered = pickle.load(f)

with open('{}/data/wordlist.pickle'.format(mod_path), mode='rb') as file:
    wordlist = pickle.load(file)

with open('{}/data/clf_NUMB.pickle'.format(mod_path), mode='rb') as file:
    clf_NUMB = pickle.load(file)

if __name__ == "__main__":
    # Store all NUMB tags from training data in NUMB_list, including SPLT-NUMB.
    tagged = tagify(NSWs, verbose=False)

    NUMB_dict = {ind: (nsw, tag) for ind, (nsw, tag) in tagged.items()
                 if tag == 'NUMB'}

    SPLT_dict = {ind: (nsw, tag) for ind, (nsw, tag) in tagged.items()
                 if tag == 'SPLT'}

    splitted = split(SPLT_dict, verbose=False)
    retagged = retagify(splitted, verbose=False)
    retagged_NUMB_dict = {ind: (nsw, tag)
                          for ind, (nsw, tag) in retagged.items()
                          if tag == 'SPLT-NUMB'}
    NUMB_dict.update(retagged_NUMB_dict)

curr_list = ['£', '$', '€', 'Y']
ampm = ['am', 'pm', 'AM', 'PM', 'a.m.', 'p.m.', 'A.M.', 'P.M.', 'pm.', 'am.']
months = ["January", "Jan", "Jan.", "February", "Feb", "Feb.",
          "March", "Mar", "Mar.", "April", "Apr", "Apr.", "May",
          "June", "Jun", "Jun.", "July", "Jul", "Jul.", "August",
          "Aug", "Aug.", "September", "Sept", "Sept.", "October",
          "Oct", "Oct.", "November", "Nov", "Nov.", "December",
          "Dec", "Dec."]
addr_words = ['Road', 'Rd.', 'Street', 'Avenue', 'Ave.']


def run_clfNUMB(dic, text, verbose=True):
    """Train classifier on training data, return dictionary with added tag.

    dic: dictionary entry where key is index of word in orig text, value
         is a tuple with the nsw and the tag (ALPHA, NUMB, SPLT-, MISC).
    The dictionary returned has the same entries with the tuple extended with
    a more specific number tag assigned to it by the classifier.
    """
    clf = clf_NUMB
    int_tag_dict = {
                    1: 'PRCT',
                    2: 'MONEY',
                    3: 'NTIME',
                    4: 'NYER',
                    5: 'NDIG',
                    6: 'NORD',
                    7: 'NUM',
                    8: 'NRANGE',
                    9: 'NTEL',
                    10: 'NDATE',
                    11: 'NADDR',
                    12: 'NSCI'
                    }
    out = {}
    for (ind, (nsw, tag)) in dic.items():
        if verbose:
            sys.stdout.write("\r{} of {} classified".format(len(out), len(dic)))
            sys.stdout.flush()
        pred_int = int(clf.predict(gen_featuresetsNUM({ind: (nsw, tag)}, text)))
        ntag = int_tag_dict[pred_int]
        out.update({ind: (nsw, tag, ntag)})
    if verbose:
        sys.stdout.write("\r{} of {} classified".format(len(out), len(dic)))
        sys.stdout.flush()
        print("\n")
    return out


def gen_featuresetsNUM(tagged_dict, text):
    """Return an array for features for each item in the input dict."""
    return np.array([give_featuresNUM(item, text)
                    for item in tagged_dict.items()],
                    dtype=float)


def give_featuresNUM(item, text):
    """Return a list of features for a dictionary item."""
    ind, nsw, tag = item[0], item[1][0], item[1][1]
    context = gen_frame(item, text)
    out = [
           adj_month(context),
           year_size(nsw) or year_context(context),
           context[1] == 'in',
           range_vs_date_slash(nsw),
           range_vs_date_hyph(nsw),
           date_vs_num(nsw),
           time_context(nsw, context),
           looks_rangey(nsw),
           nsw.isdigit() and 'SPLT' not in tag,
           nsw.isdigit() and 'SPLT' in tag,
           context[1] in ['around', 'about', 'approximately'],
           context[1] in ['over', 'under'],
           context[1] == 'on',
           context[1] == 'in',
           context[1] == 'of'
           ]
    out.extend(seed_features(item, context))
    out.extend(in_features(nsw))
    return out


def adj_month(context):
    """Return 'True' if word is adjacent to nsw."""
    if context[3] in months or context[1] in months:
        return True
    else:
        return False


def looks_rangey(nsw):
    """Return True if number fits range pattern, not date."""
    m = range_pattern.match(nsw)
    n = seed_range_pattern.match(nsw)
    if m:
        if m.group(1) and m.group(3):
            if len(m.group(1)) == 4 and len(m.group(3)) == 2:
                if m.group(1).isdigit() and m.group(3).isdigit():
                    if int(m.group(1)[-2:]) < int(m.group(3)):
                        return True
                    else:
                        return False
                elif (m or n) and not range_vs_date_hyph(nsw):
                    return True
            elif m.group(1).isdigit() and m.group(3).isdigit():
                if ((int(m.group(1)) >= int(m.group(3))) or m.group(3).startswith('0')
                or m.group(1).startswith('0')):
                    return False
                else:
                    return True
            elif (m or n) and not range_vs_date_hyph(nsw):
                return True
    elif (m or n) and not range_vs_date_hyph(nsw):
        return True
    else:
        return False


def year_size(nsw):
    """Return '1' if nsw is greater than 1800."""
    if not nsw.isdigit():
        return False
    elif 1800 < int(nsw) < 2050:
        return 1
    else:
        return 0


def year_context(context):
    adbc = ['AD', 'A.D.', 'ad', 'a.d.', 'BC', 'B.C.', 'bc', 'B.C.']
    """Return 'True' if nsw is followed by variant on 'ad' or 'bc'."""
    if context[3] in adbc or context[1] in adbc:
        return True
    else:
        return False


def range_vs_date_slash(nsw):
    """Return 'True'1' if nsw looks like a date with a slash in."""
    if '/' not in nsw:
        return False
    elif nsw.count('/') == 2:
        return 1
    elif nsw[0] == '0':
        return 1
    else:
        slash = nsw.find('/')
        first = nsw[:slash]
        second = nsw[slash+1:]
        if not (first.isdigit() and second.isdigit()):
            return False
        elif len(first) != len(second):
            return 0
        elif (int(first) > 12 and int(second) > 12 or
              int(first) > 31 or int(second) > 31):
            return 0
        elif len(first) == 1:
            return False
        elif int(second) > int(first[-len(second):]):
            return 0
        else:
            return 1


def range_vs_date_hyph(nsw):
    """Return 'True'1' if nsw looks like a date with a hyphen in."""
    if '-' not in nsw:
        return False
    elif nsw.count('-') == 2:
        return 1
    elif nsw[0] == '0':
        return 1
    else:
        hyph = nsw.find('-')
        first = nsw[:hyph]
        second = nsw[hyph+1:]
        if not (first.isdigit() and second.isdigit()):
            return False
        elif len(first) != len(second):
            return 0
        elif (int(first) > 12 and int(second) > 12 or
              int(first) > 31 or int(second) > 31):
            return 0
        elif len(first) == 1:
            return False
        elif int(second) > int(first[-len(second):]):
            return 0
        else:
            return 1


def date_vs_num(nsw):
    """Return '1' if nsw looks like a date with a '.' in."""
    if '.' not in nsw:
        return False
    elif nsw.count('.') > 1:
        return 1
    else:
        dot = nsw.find('.')
        first = nsw[:dot]
        second = nsw[dot:]
        if not (first.isdigit() and second.isdigit()):
            return False
        elif not (first.isdigit() and second.isdigit()):
            return False
        elif len(first) > 1 and first[0] == '0':
            return 1
        elif second[-1] == '0':
            return 1
        else:
            return 0


def time_context(nsw, context):
    """Return 'True' if followed by some variant of 'am' or 'pm'."""
    if '.' in nsw or ':' in nsw:
        if context[3] in ampm:
            return True
        else:
            return False
    else:
        return False


def looks_datey(nsw, context):
    """Return 'True' if nsw looks like a valid date."""
    m = date_pattern.match(nsw)
    if date_pattern.match(nsw):
        if m.group(1) and m.group(3) and m.group(5):
            return True
        elif (int(m.group(1)) <= 12 and 12 < int(m.group(3)) < 32
            or 12 < int(m.group(1)) < 32 and int(m.group(3)) <= 12):
                return True
        elif context[1] == 'on' or context[1].lower() == 'on':
            return True
        elif m.group(1).startswith('0') or m.group(3).startswith('0'):
            return True
        else:
            return False
    else:
        return False


def in_features(nsw):
    """Return list of bools if specific punctuation is in nsw."""
    out = []
    for c in [':']:
        out.append(c in nsw)
    return out


def seed_features(item, context):
    """Return a list of features equivalent to those used in the seedset."""
    ind, nsw, tag = item[0], item[1][0], item[1][1]
    out = [
           nsw.endswith('%'),
           (nsw[0] in curr_list or nsw[-3:] in ecurr_dict
            or nsw[:3] in ecurr_dict),
           ('.' in nsw
            and (context[3] in timezone_dict
             or context[3] in ampm)
             or bool(time_pattern.match(nsw))),
           ((context[1] == 'in' and len(nsw) == 4 and nsw.isdigit())
            or (nsw.endswith('s') and len(nsw) == 5
            and nsw[:2] in ['19', '20'])
            or context[1] in months and len(nsw) == 4 and nsw.isdigit()),
           (nsw.isdigit() and context[1].isalpha() and context[1].isupper()
            and len(context[1]) > 1 and context[1].lower() not in wordlist
            and len(nsw) > 1) or nsw.count('.') > 2,
           (nsw[-2:] in ['st', 'nd', 'rd', 'th']
            or ((context[1] in months or context[3] in months)
            and nsw.isdigit()
            and int(nsw) < 31)),
           bool(digit_pattern.match(nsw)),
           ((float_pattern.match(nsw) and float(nsw) > 31)
           or ((context[3] in ['million', 'billion', 'thousand']
          or (context[3] in meas_dict or context[3] in meas_dict_pl)
          and nsw.isdigit()))),
          ('/' in nsw and (context[3] in meas_dict or context[3] in meas_dict_pl
          or context[3] in meas_dict.values() or context[3] in meas_dict_pl.values())),
          looks_datey(nsw, context),
          (len(nsw) == 11 and nsw.startswith('0')) or (nsw.startswith('+44')
           and len(nsw) == 13),
          looks_rangey(nsw),
          (nsw.isdigit() and 1950 < int(nsw) < 2100
           and not (context[3] in meas_dict or context[3] in meas_dict_pl)),
          len(nsw) < 5 and context[4] in addr_words,
          bool(coord_pattern.match(nsw)) or bool(feet_pattern.match(nsw))
           ]
    return out


def create_NUMB_ex():
    """Create text file 'NUMB_examples.txt' containing NUMB nsws in context"""
    NUMB_ex = []
    for ind, (word, tag) in NUMB_dict.items():
        NUMB_ex.append(gen_frame((ind, (word, tag)), text))
    with open('NUMB_examples.txt', mode='w', encoding='utf-8') as file:
        file.write(str(NUMB_ex))


def gen_frame(dict_tup, text):
    """Return tuple containing nsw and two words either side in orig text.

    If nsw has been split, count previous parts of original 'word' as part
    of the left-context.
    If a context word is '.' and the following word is capitalised,
    replace following words in that half of the context with <END>.
    """
    ind, word, tag = dict_tup[0], dict_tup[1][0], dict_tup[1][1]
    tup = ()
    end = 0
    if isinstance(ind, int):
        for i in range(ind - 2, ind + 3):
            if i < 0:
                tup += ('<END>',)
            elif i + 2 > len(text):
                tup += (text[i],)
                rem = 5 - len(tup)
                tup += rem * ('<END>',)
                break
            elif (text[i] == '.' and
                    text[i + 1].istitle()):
                if len(tup) > 2:
                    end = 1
                    tup += ('<END>',)
                elif len(tup) == 0:
                    tup = ('<END>',)
                else:
                    leng = len(tup)
                    tup = leng * ('<END>',)
            elif end:
                tup += ('<END>',)
            else:
                tup += (text[i],)
    else:
        rind = round(ind)
        full = text[rind]
        start = ''
        end = ''
        index = full.find(word)
        start = full[:index]
        end = full[index + len(word):]
        potential_context = gen_frame((rind, (word, tag)), text)
        if start and end:
            tup = (potential_context[1],
                   start, word, end, potential_context[-2])
        elif start:
            tup = (potential_context[1],
                   start, word,
                   potential_context[-2], potential_context[-1])
        elif end:
            tup = (potential_context[0],
                   potential_context[1], word,
                   end, potential_context[-2])
        else:
            tup = (potential_context[0],
                   potential_context[1], word,
                   potential_context[-2],
                   potential_context[-1])
    return tup


def gen_seed(dic, text):
    """Return a list of the (integer) labels assigned to the seedset."""
    seedset = []
    for ind, (nsw, tag) in dic.items():
        seedset.append(seed((ind, (nsw, tag)), text))
    return seedset


def gen_feats_and_seed(dic, text):
    """Return a tuple of arrays - the first of features, the second, labels."""
    seedset = []
    featset = []
    for ind, (nsw, tag) in dic.items():
        seedset.append(seed((ind, (nsw, tag)), text))
        featset.append(give_featuresNUM((ind, (nsw, tag)), text))
    return (np.array(featset, dtype=int), np.array(seedset))


def fit_clf(dic, text):
    """Fit a Label Propogation classifier to the input dictionary."""
    model = lp()
    X, y = gen_feats_and_seed(dic, text)
    model.fit(X, y)
    return model


def fit_and_store_clf(dic, text):
    """fit a Label Propogation classifier, and store in clf_NUMB.pickle"""
    clf = fit_clf(dic, text)
    with open('{}/data/clf_NUMB.pickle'.format(mod_path), 'wb') as file:
        pickle.dump(clf, file, protocol=2)


def seed(dict_tup, text):
    """Assign a seedset label to the input tuple.

    Generate seeds for the seedset by assigning integer labels to obvious
    cases. Where there is no obvious case, '-1' is returned.
    """
    ind, nsw, tag = dict_tup[0], dict_tup[1][0], dict_tup[1][1]
    context = gen_frame((ind, (nsw, tag)), text)
    if nsw.endswith('%'):
        return 1
    elif (nsw[0] in curr_list or nsw[-3:] in ecurr_dict or
            nsw[:3] in ecurr_dict):
        return 2
    elif coord_pattern.match(nsw) or feet_pattern.match(nsw):
        return 12
    elif ('.' in nsw and
          (context[3] in timezone_dict or
           context[3] in ampm) or
          time_pattern.match(nsw)):
                return 3
    elif ((context[1] == 'in' and len(nsw) == 4 and nsw.isdigit()) or
          (nsw.endswith('s') and len(nsw) == 5 and nsw[:2] in ['19', '20']) or
          context[1] in months and len(nsw) == 4 and nsw.isdigit()):
                return 4
    elif (nsw.isdigit() and context[1].isalpha() and context[1].isupper() and
          len(context[1]) > 1 and context[1].lower() not in wordlist and
          len(nsw) > 1) or nsw.count('.') > 2:
        return 5
    elif digit_pattern.match(nsw):
        return 5
    elif (nsw[-2:] in ['st', 'nd', 'rd', 'th'] or
          ((context[1] in months or context[3] in months) and
          nsw.isdigit() and
          int(nsw) <= 31)):
            return 6
    elif ((float_pattern.match(nsw) and float(nsw) > 31) or
          ((context[3] in ['million', 'billion', 'thousand']) or
          (context[3] in meas_dict or context[3] in meas_dict_pl) and
          nsw.isdigit())):
        return 7
    elif ('/' in nsw and (context[3] in meas_dict or context[3] in meas_dict_pl
          or context[3] in meas_dict.values() or context[3] in meas_dict_pl.values())):
        return 7
    elif looks_datey(nsw, context):
        return 10
    elif looks_rangey(nsw):
        return 8
    elif (len(nsw) == 11 and nsw.startswith('0')) or (nsw.startswith('+44')
          and len(nsw) == 13):
        return 9
    elif '.' in nsw or ',' in nsw:
        return 7
    elif (nsw.isdigit() and 1950 < int(nsw) < 2100
          and not (context[3] in meas_dict or context[3] in meas_dict_pl)):
        return 4
    elif len(nsw) < 5 and context[4] in addr_words:
        return 11
    else:
        return - 1

time_pattern = re.compile('''
([0-9]{1,2})            # one or two numbers
\:                      # ':'
([0-9]{2})              # two numbers
$
''', re.VERBOSE)

float_pattern = re.compile('''
[0-9]+                  # any number of digits
\.                      # '.'
[0-9]+                  # any number of digits
$                       # end
''', re.VERBOSE)

seed_range_pattern = re.compile('''
(([0-9]{4}              # (4 digits
-                       # a hyphen
[0-9]{2}                # 2 digits
$)                      # end)
|                       # OR
([0-9]{4}               # (4 digits
/                       # a slash
[0-9]{,2}               # 1 or 2 digits
$))                     # end)
''', re.VERBOSE)

range_pattern = re.compile('''
([0-9]+                 # 1 or more digits
\.?                     # optional '.'
[0-9]*)                 # 0 or more digits
(/|-|–)                   # hyphen or slash
([0-9]+                 # same pattern as lines 1-3
\.?
[0-9]*)
$
''', re.VERBOSE)

date_pattern = re.compile('''
([0-9]{1,2})
(/)
([0-9]{1,2})
(/)?
([0-9]{2,4})?
$
''', re.VERBOSE)

coord_pattern = re.compile('''
([0-9]+
\.?
[0-9]*
°)
([0-9]+
\.?
[0-9]*
[\'|\’|′])?
([0-9]+
\.?
[0-9]*
["|″]?)?
([N|S|E|W])?
$
''', re.VERBOSE)

feet_pattern = re.compile('''
([0-9]+
[\'|\’|′])
(\.?
[0-9]+
["|″]?)?
$
''', re.VERBOSE)

digit_pattern = re.compile('''
([0-9]+
\-){2,}
[0-9]+
$
''', re.VERBOSE)



# -*- coding: utf-8 -*-

timezone_dict = {
                 "ACDT": "Australian Central Daylight Time",
                 "ACST": "Australian Central Standard Time",
                 "ACWT": "Australian Central Western Time",
                 "ADT": "Atlantic Daylight Time",
                 "ACT": "Acre Time",
                 "AEDT": "Australian Eastern Daylight Time",
                 "AEST": "Australian Eastern Standard Time",
                 "AFT": "Afghanistan Time",
                 "AKDT": "Alaska Daylight Time",
                 "AKST": "Alaska Standard Time",
                 "ALMT": "Alma-Ata Time",
                 "AMT": "Armenia Time",
                 "AMST": "Amazon Summer Time",
                 "AMST": "Armenia Summer Time",
                 "ANAT": "Anadyr Time",
                 "ANAST": "Anadyr Summer Time",
                 "AQTT": "Aqtobe Time",
                 "ART": "Argentina Time",
                 "AST": "Atlantic Standard Time",
                 "AST": "Arabia Standard Time",
                 "AWDT": "Australian Western Daylight Time",
                 "AWST": "Australian Western Standard Time",
                 "AZOT": "Azores Time",
                 "AZOST": "Azores Summer Time",
                 "AZT": "Azerbaijan Time",
                 "AZST": "Azerbaijan Summer Time",
                 "BNT": "Brunei Time",
                 "BDT": "Bangladesh Time",
                 "BOT": "Bolivia Time",
                 "BRT": "Brasilia Time",
                 "BRST": "Brasilia Summer Time",
                 "BST": "British Summer Time",
                 "BTT": "Bhutan Time",
                 "CAST": "Casey Time",
                 "CAT": "Central Africa Time",
                 "CCT": "Cocos Islands Time",
                 "CDT": "Central Daylight Time",
                 "CEDT": "Central European Daylight Time",
                 "CEST": "Central European Summer Time",
                 "CET": "Central European Time",
                 "CHADT": "Chatham Island Daylight Time",
                 "CHAST": "Chatham Island Standard Time",
                 "CHOT": "Choibalsan Time",
                 "CHOST": "Choibalsan Summer Time",
                 "CHsT": "Chamorro Standard Time",
                 "CHUT": "Chuuk Time",
                 "CIT": "Central Indonesian Time",
                 "CKT": "Cook Island Time",
                 "CLST": "Chile Summer Time",
                 "CLT": "Chile Standard Time",
                 "COT": "Columbia Time",
                 "CST": "Central Standard Time",
                 "CVT": "Cape Verde Time",
                 "CWST": "Central Western Australia Time",
                 "CXT": "Christmas Island Time",
                 "DAVT": "Davis Time",
                 "DDUT": "Dumont D 'Urville Time",
                 "DST": "Daylight Saving Time",
                 "EASST": "Eastern Island Summer Time",
                 "EAST": "Eastern Island Standard Time",
                 "EAT": "East Africa Time",
                 "ECT": "Ecuador Time",
                 "EDT": "Eastern Daylight Time",
                 "EEDT": "Eastern European Daylight Time",
                 "EEST": "Eastern European Summer Time",
                 "EET": "Eastern European Time",
                 "EGT": "Eastern Greenland Time",
                 "EGST": "Eastern Greenland Summer Time",
                 "EST": "Eastern Standard Time",
                 "EIT": "Eastern Indonesian Time",
                 "FET": "Further-Eastern European Time",
                 "FJT": "Fiji Time",
                 "FJST": "Fiji Summer Time",
                 "FKST": "Falkland Islands Summer Time",
                 "FKT": "Falkland Islands Time",
                 "FNT": "Fernando De Noronha",
                 "GALT": "Galapagos Time",
                 "GAMT": "Gambier Time",
                 "GET": "Georgia Standard Time",
                 "GFT": "French Guiana Time",
                 "GILT": "Gilbert Island Time",
                 "GMT": "Greenwich Mean Time",
                 "GST": "Gulf Standard Time",
                 "GYT": "Guyana Time",
                 "HADT": "Hawaii-Aleutian Daylight Time",
                 "HAST": "Hawaii-Aleutian Standard Time",
                 "HKT": "Hong Kong Time",
                 "HOVT": "Hovd Time",
                 "HOVST": "Hovd Summer Time",
                 "HST": "Hawaii Standard Time",
                 "ICT": "Indochina Time",
                 "IDT": "Israel Daylight Time",
                 "IOT": "Indian Chagos Time",
                 "IRDT": "Iran Daylight Time",
                 "IRKT": "Irkutsk Time",
                 "IRKST": "Irkutsk Summer Time",
                 "IRST": "Iran Standard Time",
                 "IST": "Irish Standard Time",
                 "JST": "Japan Standard Time",
                 "KGT": "Kyrgyzstan Time",
                 "KOST": "Kosrae Time",
                 "KRAT": "Krasnoyarsk Time",
                 "KRAST": "Krasnoyarsk Summer Time",
                 "KST": "Korea Standard Time",
                 "KUYT": "Kuybyshev Time",
                 "LHDT": "Lord Howe Daylight Time",
                 "LHST": "Lord Howe Standard Time",
                 "LINT": "Line Islands Time",
                 "MAGT": "Magadan Time",
                 "MAGST": "Magadan Summer Time",
                 "MART": "Marquesas Time",
                 "MAWT": "Mawson Station Time",
                 "MDT": "Mountain Daylight Time",
                 "MeST": "Metlakatla",
                 "MHT": "Marshall Islands Time",
                 "MIST": "Macquarie Island Station Time",
                 "MMT": "Myanmar Time",
                 "MSD": "Moscow Summer Time",
                 "MSK": "Moscow Standard Time",
                 "MST": "Mountain Standard Time",
                 "MUT": "Mauritius Time",
                 "MVT": "Maldives Time",
                 "MYT": "Malaysia Time",
                 "NCT": "New Caledonia Time",
                 "NDT": "Newfoundland Daylight Time",
                 "NFT": "Norfolk Time",
                 "NOVT": "Novosibirsk Time",
                 "NOVST": "Novosibirsk Summer Time",
                 "NPT": "Nepal Time",
                 "NRT": "Nauru Time",
                 "NST": "Newfoundland Standard Time",
                 "NT": "Newfoundland Time",
                 "NUT": "Niue Time",
                 "NZDT": "New Zealand Daylight Time",
                 "NZST": "New Zealand Standard Time",
                 "OMST": "Omsk Standard Time",
                 "OMSST": "Omsk Summer Time",
                 "ORAT": "Oral Time",
                 "PDT": "Pacific Daylight Time",
                 "PET": "Peru Time",
                 "PETT": "Kamchatka Time",
                 "PETST": "Kamchatka Summer Time",
                 "PGT": "Papua New Guinea Time",
                 "PHT": "Philippine Time",
                 "PHOT": "Phoenix Island Time",
                 "PKT": "Pakistan Standard Time",
                 "PMDT": "Pierre & Miquelon Daylight Time",
                 "PMST": "Pierre & Miquelon Standard Time",
                 "PONT": "Pohnpei Time",
                 "PST": "Pacific Standard Time",
                 "PWT": "Palau Time",
                 "PYT": "Paraguay Time",
                 "PYST": "Paraguay Summer Time",
                 "QYZT": "Qyzylorda Time",
                 "RET": "Reunion Time",
                 "ROTT": "Rothera",
                 "SAKT": "Sakhalin Time",
                 "SAKST": "Sakhalin Summer Time",
                 "SAMT": "Samara Time",
                 "SAST": "South Africa Standard Time",
                 "SBT": "Solomon Islands Time",
                 "SCT": "Seychelles Time",
                 "SGT": "Singapore Time",
                 "SRT": "Suriname Time",
                 "SLT": "Sri Lanka Time",
                 "SLST": "Sri Lanka Time",
                 "SRET": "Srednekolymsk Time",
                 "SST": "Samoa Standard Time",
                 "SYOT": "Syowa",
                 "TAHT": "Tahiti Time",
                 "TFT": "French Southern And Antarctic Territories Time",
                 "TJT": "Tajikistan Time",
                 "TKT": "Tokelau Time",
                 "TLT": "East Timor Time",
                 "TMT": "Turkmenistan Time",
                 "TOT": "Tonga Time",
                 "TRUT": "Truk Time",
                 "TVT": "Tuvalu Time",
                 "ULAT": "Ulaanbaatar Time",
                 "ULAST": "Ulaanbaatar Summer Time",
                 "UTC": "Coordinated Universal Time",
                 "UYST": "Uruguay Summer Time",
                 "UYT": "Uruguay Standard Time",
                 "UZT": "Uzbekistan Time",
                 "VET": "Venezuelan Standard Time",
                 "VLAT": "Vladivostok Time",
                 "VLAST": "Vladivostok Summer Time",
                 "VOLT": "Volgograd Time",
                 "VUT": "Vanuatu Time",
                 "WAKT": "Wake Island Time",
                 "WAT": "West Africa Time",
                 "WART": "West Argentina Time",
                 "WAST": "West Africa Summer Time",
                 "WDT": "Western Daylight Time",
                 "WEDT": "Western European Daylight Time",
                 "WEST": "Western European Summer Time",
                 "WET": "Western European Time",
                 "WFT": "Wallis And Futuna Time",
                 "WGT": "Western Greenland Time",
                 "WGST": "Western Greenland Summer Time",
                 "WIB": "Western Indonesian Time",
                 "WIT": "Eastern Indonesian Time",
                 "WITA": "Central Indonesian Time",
                 "WST": "Western Standard Time",
                 "WT": "Western Sahara Standard Time",
                 "YAKT": "Yakutsk Time",
                 "YAKST": "Yakutsk Summer Time",
                 "YAP": "Yap Time",
                 "YEKT": "Yekaterinburg Time",
                 "YEKST": "Yekaterinburg Summer Time",
                 }



# -*- coding: utf-8 -*-

contractions = [
                "aren't",
                "can't",
                "could've",
                "couldn't",
                "couldn't've",
                "didn't",
                "doesn't",
                "don't",
                "gonna",
                "gotta",
                "hadn't",
                "hadn't've",
                "hasn't",
                "hasn't've",
                "haven't",
                "he'd",
                "he'd'n't",
                "he'd've",
                "he'll",
                "he’s",
                "how'd",
                "how'll",
                "how's",
                "i'd",
                "i'd'n't",
                "i'd'n't've",
                "i'd've",
                "i'll",
                "i'm",
                "i've",
                "isn't",
                "it'd",
                "it'd've",
                "it'll",
                "it's",
                "let's",
                "ma'am",
                "mightn't",
                "mightn't’ve",
                "might've",
                "mustn't",
                "mustn't've",
                "must've",
                "needn't",
                "o'clock",
                "ol'",
                "oughtn't",
                "oughtn't've",
                "shan't",
                "she'd",
                "she'd'n't",
                "she'd've",
                "she'll",
                "she's",
                "should've",
                "shouldn't",
                "shouldn't've",
                "somebody'd",
                "somebody'll",
                "somebody's",
                "someone'd",
                "someone'll",
                "someone's",
                "something'll",
                "something's",
                "'sup",
                "that'll",
                "that's",
                "that'd",
                "there'd",
                "there'd'n't",
                "there'd'n't've",
                "there'd've",
                "there're",
                "there's",
                "they'd",
                "they'dn't",
                "they'dn't've",
                "they'd've",
                "they'll",
                "they're",
                "they've",
                "'tis",
                "'twas",
                "wanna",
                "wasn't",
                "we'd",
                "we'll",
                "we're",
                "we've",
                "weren't",
                "what'd",
                "what'll",
                "what're",
                "what's",
                "what've",
                "when's",
                "where'd",
                "where's",
                "where've",
                "who'd",
                "who'd've",
                "who'll",
                "who're",
                "who's",
                "who've",
                "why'd",
                "why'll",
                "why're",
                "why's",
                "won't",
                "won't've",
                "would've",
                "wouldn't",
                "wouldn't've",
                "y'all",
                "you'd",
                "you'd've",
                "you'll",
                "you're",
                "you'ren’t",
                "you've",
                "you'ven't",
                ]



# -*- coding: utf-8 -*-

from __future__ import division, print_function, unicode_literals

import pickle
from collections import defaultdict

from nltk.corpus import treebank, brown
from nltk.tag.mapping import map_tag

from normalise.data.abbrev_dict import states


def store_pos_tag_dicts():
    pos_tag_dict = defaultdict(tuple)
    tagged = treebank.tagged_sents()
    for sent in tagged:
        for tup in sent:
            if not tup[1] in pos_tag_dict[tup[0].lower()]:
                pos_tag_dict[tup[0].lower()] += (tup[1],)

    pos_tag_dict_univ = defaultdict(tuple)
    penn_tagged_univ = treebank.tagged_sents(tagset='universal')
    brown_tagged_univ = brown.tagged_sents(tagset='universal')
    for text in [penn_tagged_univ, brown_tagged_univ]:
        for sent in text:
            for tup in sent:
                if not tup[1] in pos_tag_dict_univ[tup[0].lower()]:
                    pos_tag_dict_univ[tup[0].lower()] += (tup[1],)
    for word in states.values():
        pos_tag_dict[word.lower()] += ('NNP',)
        pos_tag_dict_univ[word.lower()] += ('NOUN',)
    dicts = (pos_tag_dict, pos_tag_dict_univ)
    with open('{}/data/pos_dicts.pickle'.format(mod_path), 'wb') as file:
        pickle.dump(dicts, file, protocol=2)


def add_to_pos_dicts(pos_dict_new, pos_dict_new_univ):
    with open('{}/data/pos_dicts.pickle'.format(mod_path), mode='rb') as f:
        pos_dict, pos_dict_univ = pickle.load(f)

    for key in pos_dict_new:
        if type(pos_dict_new[key]) == tuple:
            for tag in pos_dict_new[key]:
                if tag not in pos_dict[key]:
                    pos_dict[key] = pos_dict[key] + (tag,)
        elif type(pos_dict_new[key]) == str:
            if pos_dict_new[key] not in pos_dict[key]:
                pos_dict[key] = pos_dict[key] + (pos_dict_new[key],)

    for key in pos_dict_new_univ:
        if type(pos_dict_new_univ[key]) == tuple:
            for tag in pos_dict_new_univ[key]:
                if tag not in pos_dict_univ[key]:
                    pos_dict_univ[key] = pos_dict_univ[key] + (tag,)
        elif type(pos_dict_new_univ[key]) == str:
            if pos_dict_new[key] not in pos_dict[key]:
                pos_dict_univ[key] = (pos_dict_univ[key]
                                     + (pos_dict_new_univ[key],))
    dicts = (pos_dict, pos_dict_univ)
    with open('{}/data/pos_dicts.pickle'.format(mod_path), mode='wb') as f:
        pickle.dump(dicts, f, protocol=2)
    print(dicts)


def create_pos_dict(dictionary):
    pos = defaultdict(tuple)
    pos_univ = defaultdict(tuple)
    for key in dictionary:
        if type(dictionary[key]) == tuple:
            pos[key] = dictionary[key]
            pos_univ[key] = map_tag('en-ptb', 'universal', dictionary[key])
        elif type(dictionary[key]) == str:
            pos[key] = (dictionary[key],)
            pos_univ[key] = (map_tag('en-ptb', 'universal', dictionary[key]),)
    return pos, pos_univ

title_dict = {
              'honourable': 'NNP',
              'general': 'NNP',
              'colonel': 'NNP',
              'lieutenant': 'NNP',
              'superintendent': 'NNP',
              'senior': 'NNP',
              'junior': 'NNP',
              'sir': 'NNP',
              'senator': 'NNP',
              'doctor': 'NNP',
              'duke': 'NNP',
              'captain': 'NNP'
              }
